#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
__all__ = [
    "Parallelepiped",
    "RectPrism",
    "Cube",
    "Rhombohedron",
    "Cuboctahedron",
    "Cylinder",
    "CylindricalSector",
]

from abc import abstractmethod, ABC
from copy import deepcopy
from math import cos, sin, sqrt

import numpy as np

from .transform_func import ShiftFunc, ScaleFunc, RotateFunc, ReflectFunc
from ..util.util import areEqual


class Shape(object):
    """ """

    DBL_TOL = 1e-5
    DEFAULT_ALIGNMENT = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float)

    # === STANDARD CONSTRUCTOR
    def __init__(self, Anchor, Alignment=None):
        self._Anchor = Anchor
        self._Alignment = Shape.DEFAULT_ALIGNMENT if Alignment is None else Alignment

    # === ASSERTION OF CLASS DESIGN
    def isConsistentWithDesign(self):
        """ """
        if self.Anchor is None:
            print("alpha")
            return False
        if type(self.Alignment) is not np.ndarray:
            print("A")
            return False
        if self.Alignment.shape != (3, 3):
            print("B")
            return False
        if not areEqual(np.linalg.det(self.Alignment), 1.0, Shape.DBL_TOL):
            print("C")
            return False
        for AlignmentAxis in self.Alignment:
            if not areEqual(np.linalg.norm(AlignmentAxis), 1.0, Shape.DBL_TOL):
                print("D")
                return False
        return True

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """

        Args:
            TransF:

        Returns:

        """
        if type(TransF) is ShiftFunc or type(TransF) is ScaleFunc:
            TransF.transform(self._Anchor)
        elif type(TransF) is RotateFunc or type(TransF) is ReflectFunc:
            TransF.transform(self._Anchor)
            TransF.transform(self._Alignment)
            assert self.isConsistentWithDesign()
        else:
            raise TypeError
        assert self.isConsistentWithDesign()

    def shift(self, Shift):
        """

        Args:
            Shift:

        Returns:

        """
        if type(Shift) is ShiftFunc:
            self.applyTransF(Shift)
        elif type(Shift) is np.ndarray:
            self.applyTransF(ShiftFunc(Shift))
        else:
            raise TypeError

    def scale(self, Scale, OriginOfScale=None):
        """

        Args:
            Scale: param OriginOfScale:  (Default value = None)
            OriginOfScale:  (Default value = None)

        Returns:

        """
        if type(Scale) is ScaleFunc:
            self.applyTransF(Scale)
        elif type(Scale) is np.ndarray:
            self.applyTransF(ScaleFunc(Scale, OriginOfScale))
        elif type(Scale) is float or type(Scale) is int:
            self.applyTransF(ScaleFunc(np.array([Scale, Scale, Scale], dtype=float)))
        else:
            raise TypeError

    def rotate(self, Rotation, OriginOfRotation=None):
        """

        Args:
            Rotation: param OriginOfRotation:  (Default value = None)
            OriginOfRotation:  (Default value = None)

        Returns:

        """
        if type(Rotation) is RotateFunc:
            self.applyTransF(Rotation)
        elif type(Rotation) is np.ndarray:
            self.applyTransF(RotateFunc(Rotation, OriginOfRotation))
        else:
            raise TypeError

    def reflect(self, Reflection):
        """

        Args:
            Reflection:

        Returns:

        """
        if type(Reflection) is ReflectFunc:
            self.applyTransF(Reflection)
        else:
            raise TypeError

    # === PROPERTY EVALUATION METHODS
    @abstractmethod
    def isInShape(self, P):
        """

        Args:
            P:

        Returns:

        """
        raise NotImplementedError

    __contains__ = isInShape

    @abstractmethod
    def getBounds(self):
        """ """
        raise NotImplementedError

    # === BASIC QUERY METHODS
    @property
    def Anchor(self):
        """ """
        return self._Anchor

    @property
    def Alignment(self):
        """ """
        return self._Alignment


class Polyhedron(Shape, ABC):
    """ """

    # === STANDARD CONSTRUCTOR
    def __init__(self, V, F, Anchor, Alignment=None):
        Shape.__init__(self, Anchor, Alignment)
        self._V = V
        self._F = F
        self._FacetNorms = self.__calcFacetNorms()
        self._FacetDirections = self.__calcFacetDirections()
        assert self.isConsistentWithDesign()

    # === ASSERTION OF CLASS DESIGN
    def isConsistentWithDesign(self):
        """ """
        if type(self.V) is not list:
            return False
        if len(self.V) < 4:
            return False
        if type(self.F) is not list:
            return False
        if len(self.F) < 4:
            return False
        for Facet in self.F:
            if type(Facet) is not list:
                return False
            if len(Facet) < 3:
                return False
        return Shape.isConsistentWithDesign(self)

    # === AUXILIARY METHODS
    def __calcFacetNorms(self):
        FacetNorms = []
        for Facet in self.F:
            Norm = np.cross(
                self.V[Facet[1]] - self.V[Facet[0]], self.V[Facet[2]] - self.V[Facet[0]]
            )
            Norm /= np.linalg.norm(Norm)
            FacetNorms.append(Norm)
        return FacetNorms

    def __calcFacetDirections(self):
        FacetDirections = []
        for Facet in self.F:
            assert len(Facet) >= 3
            FacetDirection = np.array([0, 0, 0], dtype=float)
            for v in Facet:
                FacetDirection += self.V[v]
            FacetDirection /= len(Facet)
            FacetDirection -= self.Anchor
            FacetDirections.append(FacetDirection)
        return FacetDirections

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """

        Args:
            TransF:

        Returns:

        """
        if type(TransF) is ShiftFunc:
            for v in self._V:
                TransF.transform(v)
        elif type(TransF) is ScaleFunc:
            for v in self._V:
                TransF.transform(v)
            for direction in self._FacetDirections:
                TransF.transform(direction)
            for norm in self._FacetNorms:
                norm /= TransF.Scale
        elif type(TransF) is RotateFunc:
            for v in self._V:
                TransF.transform(v)
            for direction in self._FacetDirections:
                TransF.transformDirection(direction)
            for norm in self._FacetNorms:
                TransF.transformDirection(norm)
        else:
            raise TypeError(
                "MatOpt does not support this transformation type. Please contact MatOpt developer for "
                "potential feature addition."
            )
        Shape.applyTransF(self, TransF)
        assert self.isConsistentWithDesign()

    # === PROPERTY EVALUATION METHODS
    def isInShape(self, P, tol=Shape.DBL_TOL):
        """

        Args:
            P: param tol:  (Default value = Shape.DBL_TOL)
            tol:  (Default value = Shape.DBL_TOL)

        Returns:

        """
        for f in range(len(self.F)):
            if not self.satisfiesFacet(P, f, tol):
                return False
        return True

    __contains__ = isInShape

    def satisfiesFacet(self, P, f, tol=Shape.DBL_TOL):
        """

        Args:
            P: param f:
            tol: Default value = Shape.DBL_TOL)
            f:

        Returns:

        """
        P01 = P - self.V[self.F[f][0]]
        return np.inner(P01, self.FacetNorms[f]) < tol

    def getBounds(self):
        """ """
        MinP = deepcopy(self.V[0])
        MaxP = deepcopy(self.V[0])
        for v in self.V:
            MinP = np.minimum(MinP, v)
            MaxP = np.maximum(MaxP, v)
        return MinP, MaxP

    # === BASIC QUERY METHODS
    @property
    def V(self):
        """ """
        return self._V

    @property
    def F(self):
        """ """
        return self._F

    @property
    def FacetNorms(self):
        """ """
        return self._FacetNorms

    @property
    def FacetDirections(self):
        """ """
        return self._FacetDirections


class Cuboctahedron(Polyhedron, ABC):
    """ """

    # === STANDARD CONSTRUCTOR
    def __init__(self, R, Center=None):
        self._R = sqrt(2)
        V = [
            np.array([1, 0, 1], dtype=float),
            np.array([0, -1, 1], dtype=float),
            np.array([-1, 0, 1], dtype=float),
            np.array([0, 1, 1], dtype=float),
            np.array([1, 1, 0], dtype=float),
            np.array([1, -1, 0], dtype=float),
            np.array([-1, -1, 0], dtype=float),
            np.array([-1, 1, 0], dtype=float),
            np.array([1, 0, -1], dtype=float),
            np.array([0, -1, -1], dtype=float),
            np.array([-1, 0, -1], dtype=float),
            np.array([0, 1, -1], dtype=float),
        ]
        F = [
            [0, 3, 2, 1],
            [1, 6, 9, 5],
            [2, 7, 10, 6],
            [3, 4, 11, 7],
            [0, 5, 8, 4],
            [8, 9, 10, 11],
            [1, 2, 6],
            [2, 3, 7],
            [0, 4, 3],
            [0, 1, 5],
            [6, 10, 9],
            [7, 11, 10],
            [4, 8, 11],
            [5, 9, 8],
        ]
        Polyhedron.__init__(self, V, F, np.array([0, 0, 0], dtype=float))
        self.scale(R / sqrt(2))
        if Center is not None:
            self.shift(Center)

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """

        Args:
            TransF:

        Returns:

        """
        if isinstance(TransF, ScaleFunc):
            if TransF.isIsometric:
                self._R *= TransF.Scale[0]
            else:
                raise ValueError(
                    "Cuboctahedron applyTransF: Can only scale isometrically"
                )
        Polyhedron.applyTransF(self, TransF)
        assert self.isConsistentWithDesign()


class Parallelepiped(Polyhedron, ABC):
    """ """

    # === STANDARD CONSTRUCTOR
    def __init__(self, Vx, Vy, Vz, BotBackLeftCorner=None):
        self._Vx = Vx
        self._Vy = Vy
        self._Vz = Vz
        if np.inner(np.cross(Vx, Vy), Vz) < Parallelepiped.DBL_TOL:
            raise ValueError("Vx,Vy,Vz must have a positive box product.")
        if BotBackLeftCorner is None:
            BotBackLeftCorner = np.array([0, 0, 0], dtype=float)

        V = [
            BotBackLeftCorner,
            BotBackLeftCorner + Vx,
            BotBackLeftCorner + Vy,
            BotBackLeftCorner + Vz,
            BotBackLeftCorner + Vx + Vy,
            BotBackLeftCorner + Vx + Vz,
            BotBackLeftCorner + Vy + Vz,
            BotBackLeftCorner + Vx + Vy + Vz,
        ]
        F = [
            [0, 1, 5, 3],
            [0, 3, 6, 2],
            [2, 6, 7, 4],
            [1, 4, 7, 5],
            [3, 5, 7, 6],
            [0, 2, 4, 1],
        ]
        Polyhedron.__init__(self, V, F, Anchor=deepcopy(BotBackLeftCorner))

    # === CONSTRUCTOR - From edge lengths and angles
    @classmethod
    def fromEdgesAndAngles(cls, A, B, C, alpha, beta, gamma, BotBackLeftCorner=None):
        """

        Args:
            A: param B:
            C: param alpha:
            beta: param gamma:
            BotBackLeftCorner: Default value = None)
            B:
            alpha:
            gamma:

        Returns:

        """
        Vx = np.array([A, 0, 0])
        Vy = np.array([B * np.cos(gamma), B * np.sin(gamma), 0])
        zcomp = np.sqrt(1 - pow(np.cos(alpha), 2) - pow(np.cos(beta), 2))
        Vz = np.array([C * np.cos(alpha), C * np.cos(beta), C * zcomp])
        return cls(Vx, Vy, Vz, BotBackLeftCorner)

    # === CONSTRUCTOR - From POSCAR file header information
    @classmethod
    def fromPOSCAR(cls, filename):
        """

        Args:
            filename:

        Returns:

        """
        with open(filename, "r") as infile:
            CommentLine = infile.readline()
            GSLine = infile.readline().split()
            GS = float(GSLine[0])
            VxLine = infile.readline().split()
            Vx = GS * np.array(
                [float(VxLine[0]), float(VxLine[1]), float(VxLine[2])], dtype=float
            )
            VyLine = infile.readline().split()
            Vy = GS * np.array(
                [float(VyLine[0]), float(VyLine[1]), float(VyLine[2])], dtype=float
            )
            VzLine = infile.readline().split()
            Vz = GS * np.array(
                [float(VzLine[0]), float(VzLine[1]), float(VzLine[2])], dtype=float
            )
        return cls(Vx, Vy, Vz)

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """

        Args:
            TransF:

        Returns:

        """
        if isinstance(TransF, ScaleFunc):
            self._Vx *= TransF.Scale[0]
            self._Vy *= TransF.Scale[1]
            self._Vz *= TransF.Scale[2]
        Polyhedron.applyTransF(self, TransF)
        assert self.isConsistentWithDesign()

        # === PROPERTY EVALUATION METHODS

    def getFractionalCoords(self, P, blnRelativeToCenter=False):
        """

        Args:
            P: param blnRelativeToCenter:  (Default value = False)
            blnRelativeToCenter:  (Default value = False)

        Returns:

        """
        FracOrigin = self.getCenter() if blnRelativeToCenter else self.Anchor
        Omega = self.getVolume()
        fracX = 1.0 / Omega * np.inner(P - FracOrigin, np.cross(self.Vy, self.Vz))
        fracY = 1.0 / Omega * np.inner(P - FracOrigin, np.cross(self.Vz, self.Vx))
        fracZ = 1.0 / Omega * np.inner(P - FracOrigin, np.cross(self.Vx, self.Vy))
        return np.array([fracX, fracY, fracZ])

    def getCenter(self):
        """ """
        return sum(self.V) / len(self.V)

    def getVolume(self):
        """ """
        return np.dot(np.cross(self.Vx, self.Vy), self.Vz)

    # === BASIC QUERY METHODS
    @property
    def Vx(self):
        """ """
        return self._Vx

    @property
    def Vy(self):
        """ """
        return self._Vy

    @property
    def Vz(self):
        """ """
        return self._Vz

    @property
    def isUnit(self):
        """ """
        if not areEqual(np.linalg.norm(self.Vx), 1.0, Shape.DBL_TOL):
            return False
        if not areEqual(np.linalg.norm(self.Vy), 1.0, Shape.DBL_TOL):
            return False
        if not areEqual(np.linalg.norm(self.Vz), 1.0, Shape.DBL_TOL):
            return False
        return True


class Rhombohedron(Parallelepiped, ABC):
    """ """

    # === STANDARD CONSTRUCTOR
    def __init__(self, L, Alpha, BotBackLeftCorner=None):
        self._L = L
        self._Alpha = Alpha
        Lx = np.array([L, 0, 0])
        Ly = np.array([L * cos(Alpha), L * sin(Alpha), 0])
        Lz = np.array(
            [
                L * cos(Alpha),
                L * (cos(Alpha) - cos(Alpha) ** 2) / sin(Alpha),
                L * sqrt(1 - 3 * cos(Alpha) ** 2 + 2 * cos(Alpha) ** 3) / sin(Alpha),
            ]
        )
        Parallelepiped.__init__(self, Lx, Ly, Lz, BotBackLeftCorner)

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """

        Args:
            TransF:

        Returns:

        """
        if isinstance(TransF, ScaleFunc):
            if TransF.isIsometric:
                self._L *= TransF.Scale[0]
            else:
                raise ValueError(
                    "Rhombohedron applyTransF: Can only scale isometrically"
                )
        Polyhedron.applyTransF(self, TransF)
        assert self.isConsistentWithDesign()

        # === PROPERTY EVALUATION METHODS

    @property
    def L(self):
        """ """
        return self._L

    @property
    def Alpha(self):
        """ """
        return self._Alpha


class RectPrism(Parallelepiped, ABC):
    """ """

    # === STANDARD CONSTRUCTOR
    def __init__(self, Lx, Ly, Lz, BotBackLeftCorner=None):
        self._Lx = Lx
        self._Ly = Ly
        self._Lz = Lz
        Parallelepiped.__init__(
            self,
            np.array([Lx, 0, 0], dtype=float),
            np.array([0, Ly, 0], dtype=float),
            np.array([0, 0, Lz], dtype=float),
            BotBackLeftCorner,
        )

    # === CONSTRUCTOR - Bounding box of a list of points
    @classmethod
    def fromPointsBBox(cls, Pts):
        """

        Args:
            Pts:

        Returns:

        """
        # import code; code.interact(local=dict(locals(),**globals()));
        minX = min(_[0] for _ in Pts)
        maxX = max(_[0] for _ in Pts)
        minY = min(_[1] for _ in Pts)
        maxY = max(_[1] for _ in Pts)
        minZ = min(_[2] for _ in Pts)
        maxZ = max(_[2] for _ in Pts)
        return cls(maxX - minX, maxY - minY, maxZ - minZ, np.array([minX, minY, minZ]))

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """

        Args:
            TransF:

        Returns:

        """
        if isinstance(TransF, ScaleFunc):
            self._Lx *= TransF.Scale[0]
            self._Ly *= TransF.Scale[1]
            self._Lz *= TransF.Scale[2]
        Parallelepiped.applyTransF(self, TransF)
        assert self.isConsistentWithDesign()

        # === BASIC QUERY METHODS

    @property
    def Lx(self):
        """ """
        return self._Lx

    @property
    def Ly(self):
        """ """
        return self._Ly

    @property
    def Lz(self):
        """ """
        return self._Lz


class Cube(RectPrism, ABC):
    """ """

    # === STANDARD CONSTRUCTOR
    def __init__(self, L, BotBackLeftCorner=None):
        self._L = L
        RectPrism.__init__(self, L, L, L, BotBackLeftCorner)

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """

        Args:
            TransF:

        Returns:

        """
        if isinstance(TransF, ScaleFunc):
            if TransF.isIsometric:
                self._L *= TransF.Scale[0]
            else:
                raise ValueError("Cube applyTransF: Can only scale isometrically")
        RectPrism.applyTransF(self, TransF)
        assert self.isConsistentWithDesign()

        # === BASIC QUERY METHODS

    @property
    def L(self):
        """ """
        return self._L


class Cylinder(Shape, ABC):
    """Object to create, manipulate and utilize Cylinder shapes."""

    # === STANDARD CONSTRUCTOR
    def __init__(self, Po, R, H, Vh=np.array([0, 0, 1], dtype=float), Alignment=None):
        Shape.__init__(self, Po, Alignment)
        self._R = R
        self._H = H
        self._Vh = Vh * self._H / np.linalg.norm(Vh)
        assert self.isConsistentWithDesign()

    # === ASSERTION OF CLASS DESIGN
    def isConsistentWithDesign(self):
        if len(self.Vh) != 3:
            return False
        return self.R > 0 and self.H > 0 and Shape.isConsistentWithDesign(self)

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """Apply the input transform function.

        Args:
            TransF: Transform function.
        """
        if type(TransF) is ScaleFunc:
            if areEqual(TransF.Scale[0], TransF.Scale[1], Shape.DBL_TOL):
                self._R *= TransF.Scale[0]
                self._H *= TransF.Scale[2]
                TransF.transform(self._Vh)
            else:
                raise ValueError(
                    "The scaling ratio along the radius directions are not consistent"
                )
        elif (type(TransF) is RotateFunc) or (type(TransF) is ReflectFunc):
            TransF.transform(self._Vh)
        elif not (type(TransF) is ShiftFunc):
            raise TypeError(
                "MatOpt does not support this transformation type. Please contact MatOpt developer for "
                "potential feature addition."
            )
        Shape.applyTransF(self, TransF)
        assert self.isConsistentWithDesign()

    # === PROPERTY EVALUATION METHODS
    def isInShape(self, P, tol=Shape.DBL_TOL):
        """Check if a point is inside the Cylinder shape within certain tolerance.

        Args:
            P: Input point
            tol: Tolerance (Default value = Shape.DBL_TOL)

        Returns:
            True/False
        """
        OP = P - self.Anchor
        projH = np.dot(OP, self.Vh) / self.H
        if (projH > -tol) and (projH < self.H + tol):
            normOP = np.linalg.norm(OP)
            projR = sqrt(normOP**2 - projH**2 + tol)
            return projR < self.R + tol
        else:
            return False

    __contains__ = isInShape

    def getBounds(self):
        """Get the bounds of each coordinate in the form of an artificial point."""
        MinP = (
            np.minimum(np.zeros(3, dtype=float), self.Vh)
            + self.Anchor
            - np.ones(3, dtype=float) * self.R
        )
        MaxP = (
            np.maximum(np.zeros(3, dtype=float), self.Vh)
            + self.Anchor
            + np.ones(3, dtype=float) * self.R
        )
        return MinP, MaxP

    # === BASIC QUERY METHODS
    @property
    def R(self):
        """Radius."""
        return self._R

    @property
    def H(self):
        """Height."""
        return self._H

    @property
    def Vh(self):
        """Center axis vector."""
        return self._Vh


class CylindricalSector(Shape, ABC):
    """Object to create, manipulate and utilize Cylinder shapes."""

    # === STANDARD CONSTRUCTOR
    def __init__(self, Po, R, H, Va, Vb, Vh, Alignment=None):
        Shape.__init__(self, Po, Alignment)
        self._R = R
        self._H = H
        self._Vh = Vh * self._H / np.linalg.norm(Vh)
        self._Va = Va * self._R / np.linalg.norm(Va)
        self._Vb = Vb * self._R / np.linalg.norm(Vb)
        self._Norms = self.setNorms()
        assert self.isConsistentWithDesign()

    # === ASSERTION OF CLASS DESIGN
    def isConsistentWithDesign(self):
        if (
            self.R > 0
            and self.H > 0
            and len(self.Vh) == 3
            and len(self.Va) == 3
            and len(self.Vb) == 3
            and len(self.Norms) == 3
        ):
            if areEqual(np.dot(self.Va, self.Vh), 0.0, Shape.DBL_TOL) and areEqual(
                np.dot(self.Vb, self.Vh), 0.0, Shape.DBL_TOL
            ):
                return Shape.isConsistentWithDesign(self)
        return False

    # === AUXILIARY METHODS
    def setNorms(self):
        return [
            np.cross(self._Vb, self._Va),
            np.cross(self._Va, self._Vh),
            np.cross(self._Vh, self._Vb),
        ]

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        """Apply the input transform function.

        Args:
            TransF: Transform function.
        """
        if type(TransF) is ScaleFunc:
            if areEqual(TransF.Scale[0], TransF.Scale[1], Shape.DBL_TOL):
                self._R *= TransF.Scale[0]
                self._H *= TransF.Scale[2]
                TransF.transform(self._Vh)
            else:
                raise ValueError(
                    "The scaling ratio along the radius directions are not consistent"
                )
        elif (type(TransF) is RotateFunc) or (type(TransF) is ReflectFunc):
            TransF.transform(self._Va)
            TransF.transform(self._Vb)
            TransF.transform(self._Vh)
            self._Norms = self.setNorms()
        elif not (type(TransF) is ShiftFunc):
            raise TypeError(
                "MatOpt does not support this transformation type. Please contact MatOpt developer for "
                "potential feature addition."
            )
        Shape.applyTransF(self, TransF)
        assert self.isConsistentWithDesign()

    # === PROPERTY EVALUATION METHODS
    def isInShape(self, P, tol=Shape.DBL_TOL):
        """Check if a point is inside the Cylinder shape within certain tolerance.

        Args:
            P: Input point
            tol: Tolerance (Default value = Shape.DBL_TOL)

        Returns:
            True/False
        """
        OP = P - self.Anchor
        for n in self.Norms:
            if np.dot(n, OP) > tol:
                return False
        if np.dot(-self.Norms[0], OP - self.Vh) > tol:
            return False
        projH = np.linalg.norm(np.dot(OP, self.Vh)) / self.H
        normP = np.linalg.norm(OP)
        if sqrt(normP**2 - projH**2 + tol) > self.R + tol:
            return False
        return True

    __contains__ = isInShape

    # === BASIC QUERY METHODS
    @property
    def R(self):
        """Radius."""
        return self._R

    @property
    def H(self):
        """Height."""
        return self._H

    @property
    def Vh(self):
        """Center axis vector."""
        return self._Vh

    @property
    def Va(self):
        """Left edge vector."""
        return self._Va

    @property
    def Vb(self):
        """Right edge vector."""
        return self._Vb

    @property
    def Norms(self):
        """Norms."""
        return self._Norms



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Predefined expression for Helmholtz EoS functions
"""

__author__ = "John Eslick"

import pyomo.environ as pyo


def phi_ideal_expressions_type03(model, parameters):
    """Type03 expression for the ideal part of dimensionless Helmholtz free energy

    Args:
        model (Block): Pyomo model
        parameters (dict): Main parameters dictionary

    Returns:
        dict: Expressions for ideal part of Helmholtz free energy
    """
    last_term = parameters["eos"]["last_term_ideal"]
    n0 = parameters["eos"]["n0"]
    g0 = parameters["eos"]["g0"]
    rng1 = range(4, last_term + 1)
    return {
        "phii": pyo.log(model.delta)
        + n0[1]
        + n0[2] * model.tau
        + n0[3] * pyo.log(model.tau)
        + sum(n0[i] * model.tau ** g0[i] for i in rng1),
        "phii_d": 1.0 / model.delta,
        "phii_dd": -1.0 / model.delta**2,
        "phii_t": n0[2]
        + n0[3] / model.tau
        + sum(n0[i] * g0[i] * model.tau ** (g0[i] - 1) for i in rng1),
        "phii_tt": -n0[3] / model.tau**2
        + sum(n0[i] * g0[i] * (g0[i] - 1) * model.tau ** (g0[i] - 2) for i in rng1),
        "phii_dt": 0,
    }



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Eucken equation for thermal conductivity from
Properties of Gases and Liquids, 5th Ed., 10-3-1

Authors: Douglas Allan
"""

import pytest
import types

from pyomo.environ import ConcreteModel, Block, value, Var, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.pure.Eucken import Eucken
from idaes.core.util.misc import add_object_reference
from idaes.core.base.property_meta import PropertyClassMetadata
from idaes.core.util.exceptions import ConfigurationError


def construct_dummy_model(param_dict):
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    # Parameters for sulfur dioxide, from Properties of Gases and Liquids 5th ed. Appendix B.
    m.params.config.parameter_data = {
        "f_int_eucken": (1.61, pyunits.dimensionless),
    }

    # Also need to dummy configblock on the model for the test
    m.config = ConfigBlock(implicit=True)

    m.meta_object = PropertyClassMetadata()
    m.meta_object._default_units.set_units(
        temperature=pyunits.K,
        mass=pyunits.kg,
        length=pyunits.m,
        time=pyunits.s,
        amount=pyunits.mol,
    )

    def get_metadata(self):
        return m.meta_object

    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Create variables that should exist on param block
    m.params.mw = Var(initialize=param_dict["mw"], units=pyunits.kg / pyunits.mol)
    # Hacking cp on the fake component object
    m.params.cp_mol_pure = Var(initialize=1, units=pyunits.J / pyunits.mol / pyunits.K)
    m.params.config.cp_mol_ig_comp = Block()

    def return_expression(b, cobj, T):
        return cobj.cp_mol_pure

    m.params.config.cp_mol_ig_comp.return_expression = return_expression

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=298, units=pyunits.K)
    # Have to trick the block into considering "params" as the component name
    m.props[1]._visc_d_phase_comp = Var(
        ["Vap"], ["params"], initialize=1e-7, units=pyunits.Pa * pyunits.s
    )
    return m


@pytest.mark.unit
def test_therm_cond_phase_comp_acetone():
    m = construct_dummy_model({"mw": 58.080e-3})

    Eucken.therm_cond_phase_comp.build_parameters(m.params, "Vap")

    assert isinstance(m.params.f_int_eucken, Var)
    assert value(m.params.f_int_eucken) == pytest.approx(1.61, rel=1e-12)

    expr = Eucken.therm_cond_phase_comp.return_expression(
        m.props[1], m.params, "Vap", m.props[1].temperature
    )

    # Pulled from Table 10-2, Properties of Gases and Liquids 5th Ed.
    T_list = [353, 393, 457]  # Gas temperature in K
    therm_cond_list = [
        15.7e-3,
        19.4e-3,
        24.7e-3,
    ]  # Experimental thermal conductivity in W/(m K)
    visc_list = [90.0e-7, 100e-7, 114.5e-7]  # Viscosities in Pascal seconds
    Cv_list = [77.9, 84.2, 96.1]  # Constant *volume* heat capacity in J/(mol K)
    f_int_list = [1, 1.32, 1.15]
    err_list = [
        [-4.7, -8.7, -8.4],  # Eucken percent error
        [16, 12, 13],  # Modified Eucken percent error
        [5.1, 0.9, 1.7],
    ]  # Stiel and Thodos percent error
    for i in range(3):
        m.props[1].temperature.value = T_list[i]
        m.props[1]._visc_d_phase_comp["Vap", "params"].value = visc_list[i]
        m.params.cp_mol_pure.value = Cv_list[i] + 8.314
        for j in range(3):
            m.params.f_int_eucken.value = f_int_list[j]
            err = 100 * value(expr / therm_cond_list[i] - 1)
            # Obtain percent error, then compare to that tabulated
            # in Table 10-2
            assert err == pytest.approx(err_list[j][i], abs=0.3)

    assert_units_equivalent(expr, pyunits.W / pyunits.m / pyunits.K)


@pytest.mark.unit
def test_therm_cond_phase_comp_CO2():
    m = construct_dummy_model({"mw": 44.009e-3})

    Eucken.therm_cond_phase_comp.build_parameters(m.params, "Vap")

    assert isinstance(m.params.f_int_eucken, Var)
    assert value(m.params.f_int_eucken) == pytest.approx(1.61, rel=1e-12)

    expr = Eucken.therm_cond_phase_comp.return_expression(
        m.props[1], m.params, "Vap", m.props[1].temperature
    )

    # Pulled from Table 10-2, Properties of Gases and Liquids 5th Ed.
    T_list = [200, 300, 473, 598, 1273]  # Gas temperature in K
    therm_cond_list = [
        9.51e-3,
        16.7e-3,
        28.4e-3,
        37.9e-3,
        81.7e-3,
    ]  # Experimental thermal conductivity in W/(m K)
    visc_list = [
        101.5e-7,
        149.5e-7,
        225.0e-7,
        272.8e-7,
        465.1e-7,
    ]  # Viscosities in Pascal seconds
    Cv_list = [
        24.7,
        28.9,
        35.6,
        39.5,
        48.8,
    ]  # Constant *volume* heat capacity in J/(mol K)
    f_int_list = [1, 1.32, 1.15]
    err_list = [
        [5.3, -3.2, -2.2, -4.8, -13],  # Eucken percent error
        [15, 7.5, 11, 9.3, 2.4],  # Modified Eucken percent error
        [9.8, 1.9, 4.1, 1.9, -5.6],
    ]  # Stiel and Thodos percent error
    for i in range(5):
        m.props[1].temperature.value = T_list[i]
        m.props[1]._visc_d_phase_comp["Vap", "params"].value = visc_list[i]
        m.params.cp_mol_pure.value = Cv_list[i] + 8.314
        for j in range(3):
            m.params.f_int_eucken.value = f_int_list[j]
            err = 100 * value(expr / therm_cond_list[i] - 1)
            # Obtain percent error, then compare to that tabulated
            # in Table 10-2
            assert err == pytest.approx(err_list[j][i], abs=0.5)

    assert_units_equivalent(expr, pyunits.W / pyunits.m / pyunits.K)


@pytest.mark.unit
def test_therm_cond_phase_comp_ammonia():
    m = construct_dummy_model({"mw": 17.031e-3})

    Eucken.therm_cond_phase_comp.build_parameters(m.params, "Vap")

    assert isinstance(m.params.f_int_eucken, Var)
    assert value(m.params.f_int_eucken) == pytest.approx(1.61, rel=1e-12)

    expr = Eucken.therm_cond_phase_comp.return_expression(
        m.props[1], m.params, "Vap", m.props[1].temperature
    )

    # Pulled from Table 10-2, Properties of Gases and Liquids 5th Ed.
    T_list = [213, 273]  # Gas temperature in K
    therm_cond_list = [16.5e-3, 21.9e-3]  # Experimental thermal conductivity in W/(m K)
    visc_list = [73.2e-7, 90.6e-7]  # Viscosities in Pascal seconds
    Cv_list = [25.4, 26.7]  # Constant *volume* heat capacity in J/(mol K)
    f_int_list = [1, 1.32, 1.15]
    err_list = [
        [15, 10],  # Eucken percent error
        [26, 21],  # Modified Eucken percent error
        [20, 16],
    ]  # Stiel and Thodos percent error
    for i in range(2):
        m.props[1].temperature.value = T_list[i]
        m.props[1]._visc_d_phase_comp["Vap", "params"].value = visc_list[i]
        m.params.cp_mol_pure.value = Cv_list[i] + 8.314
        for j in range(3):
            m.params.f_int_eucken.value = f_int_list[j]
            err = 100 * value(expr / therm_cond_list[i] - 1)
            # Obtain percent error, then compare to that tabulated
            # in Table 10-2
            assert err == pytest.approx(err_list[j][i], abs=0.6)

    assert_units_equivalent(expr, pyunits.W / pyunits.m / pyunits.K)


@pytest.mark.unit
def test_no_cp_ig_error():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    # Parameters for sulfur dioxide, from Properties of Gases and Liquids 5th ed. Appendix B.
    m.params.config.parameter_data = {
        "f_int_eucken": (1.61, pyunits.dimensionless),
    }

    # Also need to dummy configblock on the model for the test
    m.config = ConfigBlock(implicit=True)

    m.meta_object = PropertyClassMetadata()
    m.meta_object._default_units.set_units(
        temperature=pyunits.K,
        mass=pyunits.kg,
        length=pyunits.m,
        time=pyunits.s,
        amount=pyunits.mol,
    )

    def get_metadata(self):
        return m.meta_object

    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Create variables that should exist on param block
    m.params.mw = Var(initialize=2.0, units=pyunits.kg / pyunits.mol)
    # Hacking cp on the fake component object

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=298, units=pyunits.K)
    # Have to trick the block into considering "params" as the component name
    m.props[1].visc_d_phase_comp = Var(
        ["Vap"], ["params"], initialize=1e-7, units=pyunits.Pa * pyunits.s
    )

    Eucken.therm_cond_phase_comp.build_parameters(m.params, "Vap")

    with pytest.raises(
        ConfigurationError,
        match="Cannot find method to calculate cp_mol_ig_comp for component params.",
    ):
        expr = Eucken.therm_cond_phase_comp.return_expression(
            m.props[1], m.params, "Vap", m.props[1].temperature
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Cubic equation of state methods

Author: Andrew Lee
"""
import pytest
from sys import modules

from pyomo.environ import (
    assert_optimal_termination,
    ConcreteModel,
    Constraint,
    Expression,
    log,
    sqrt,
    value,
    Var,
    units as pyunits,
)
from pyomo.core.expr.numeric_expr import ExternalFunctionExpression
from pyomo.util.check_units import assert_units_consistent

from idaes.core import declare_process_block_class, LiquidPhase, VaporPhase, SolidPhase
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
    GenericParameterData,
    ModularPropertiesInitializer,
)
from idaes.core.util.exceptions import PropertyNotSupportedError, ConfigurationError
from idaes.core.util.constants import Constants as const
from idaes.models.properties.modular_properties.eos.ceos import (
    cubic_roots_available,
    calculate_equilibrium_cubic_coefficients,
    EoS_param,
)
from idaes.core.solvers import get_solver
from idaes.core.initialization.initializer_base import InitializationStatus
from idaes.models.properties.modular_properties.state_definitions import FTPx


# Dummy method for property method calls
def dummy_call(b, j, T):
    return 42


# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


# Dummy methods for dummied submodules
class dummy_pe:
    def return_expression(b, *args):
        # Return a dummy expression for the constraint
        return b.temperature == 100


def phase_equil(b, *args):
    pass


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True

    def parameters(self):
        self.parameters_set = True


def define_state(b):
    # Add common variables
    b.pressure = Var(initialize=101325, units=pyunits.Pa)
    b.temperature = Var(initialize=300, units=pyunits.K)
    b.mole_frac_phase_comp = Var(
        b.params.phase_list, b.params.component_list, initialize=0.5
    )
    b.log_mole_frac_phase_comp = Var(
        b.params.phase_list, b.params.component_list, initialize=0
    )

    # Set mole fractions for each phase so they can be distinguished
    if "Sol" not in b.params.phase_list:
        b.mole_frac_phase_comp["Vap", "a"].value = 0.1
        b.mole_frac_phase_comp["Vap", "b"].value = 0.2
        b.mole_frac_phase_comp["Vap", "c"].value = 0.7
        b.mole_frac_phase_comp["Liq", "a"].value = 0.6
        b.mole_frac_phase_comp["Liq", "b"].value = 0.3
        b.mole_frac_phase_comp["Liq", "c"].value = 0.1

        for p in ["Liq", "Vap"]:
            for j in "abc":
                b.log_mole_frac_phase_comp[p, j].value = log(
                    b.mole_frac_phase_comp[p, j].value
                )


@pytest.fixture()
def m():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(
        components={
            "a": {
                "phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                "parameter_data": {
                    "omega": 0.1,
                    "pressure_crit": 100000.0,
                    "temperature_crit": 100,
                },
            },
            "b": {
                "phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                "parameter_data": {
                    "omega": 0.2,
                    "pressure_crit": 200000.0,
                    "temperature_crit": 200,
                },
            },
            "c": {
                "phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                "parameter_data": {
                    "omega": 0.3,
                    "pressure_crit": 300000.0,
                    "temperature_crit": 300,
                },
            },
        },
        phases={
            "Vap": {
                "type": VaporPhase,
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
            "Liq": {
                "type": LiquidPhase,
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
        },
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        state_definition=modules[__name__],
        pressure_ref=100000.0,
        temperature_ref=300,
        phases_in_equilibrium=[("Vap", "Liq")],
        phase_equilibrium_state={("Vap", "Liq"): modules[__name__]},
        parameter_data={
            "PR_kappa": {
                ("a", "a"): 0.0,
                ("a", "b"): 0.0,
                ("a", "c"): 0.0,
                ("b", "a"): 0.0,
                ("b", "b"): 0.0,
                ("b", "c"): 0.0,
                ("c", "a"): 0.0,
                ("c", "b"): 0.0,
                ("c", "c"): 0.0,
            }
        },
    )

    m.props = m.params.state_block_class(
        [1], defined_state=False, parameters=m.params, has_phase_equilibrium=True
    )

    # Set a distinct value for _teq so it can be distinguished from temperature
    m.props[1]._teq[("Vap", "Liq")].value = 100

    m.props[1].energy_internal_mol_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=1
    )
    m.props[1].enth_mol_phase_comp = Var(m.params.phase_list, m.params.component_list)
    m.props[1].vol_mol_phase_comp = Var(m.params.phase_list, m.params.component_list)
    m.props[1].entr_mol_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=1
    )

    m.props[1].enth_mol_phase = Var(m.params.phase_list)
    m.props[1].entr_mol_phase = Var(m.params.phase_list)

    m.props[1].dens_mol_phase = Var(m.params.phase_list)
    m.props[1].mw_phase = Var(m.params.phase_list)

    for j in m.params.component_list:
        m.params.config.include_enthalpy_of_formation = False
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

        m.params.get_component(j).config.entr_mol_liq_comp = dummy_call
        m.params.get_component(j).config.entr_mol_ig_comp = dummy_call

    return m


# Expected values for A and B
Al = 0.0419062
Bl = 0.0262770
Av = 0.1180005
Bv = 0.0262769
Zl = 0.9870125
Zv = 0.9067390

Al_eq = 0.8325554
Bl_eq = 0.0788309
Av_eq = 1.9805792
Bv_eq = 0.0788309

Hl = -212.211
Hv = -779.305
Ul = Hl - 8.314 * 300 * (Zl - 1)
Uv = Hv - 8.314 * 300 * (Zv - 1)


@pytest.mark.unit
def test_wrong_phase():
    m = ConcreteModel()
    with pytest.raises(PropertyNotSupportedError):
        m.params = DummyParameterBlock(
            components={
                "a": {
                    "phase_equilibrium_form": {("Sol", "Liq"): dummy_pe},
                    "parameter_data": {
                        "omega": 0.1,
                        "pressure_crit": 100000.0,
                        "temperature_crit": 100,
                    },
                }
            },
            phases={
                "Sol": {
                    "type": SolidPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                }
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            phases_in_equilibrium=[("Vap", "Liq")],
            phase_equilibrium_state={("Vap", "Liq"): modules[__name__]},
            parameter_data={"PR_kappa": {("a", "a"): 0.0}},
        )


@pytest.mark.unit
def test_unknown_eos_option():
    m = ConcreteModel()
    with pytest.raises(ConfigurationError):
        m.params = DummyParameterBlock(
            components={
                "a": {
                    "phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                    "parameter_data": {
                        "omega": 0.1,
                        "pressure_crit": 100000.0,
                        "temperature_crit": 100,
                    },
                }
            },
            phases={
                "Vap": {
                    "type": VaporPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": "PR"},
                }
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            phases_in_equilibrium=[("Vap", "Liq")],
            phase_equilibrium_state={("Vap", "Liq"): modules[__name__]},
            parameter_data={"PR_kappa": {("a", "a"): 0.0}},
        )


@pytest.mark.unit
def test_mixing_rule_fail():
    m = ConcreteModel()

    # Dummy params block
    with pytest.raises(ValueError):
        m.params = DummyParameterBlock(
            components={
                "a": {
                    "phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                    "parameter_data": {
                        "omega": 0.1,
                        "pressure_crit": 100000.0,
                        "temperature_crit": 100,
                    },
                },
                "b": {
                    "phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                    "parameter_data": {
                        "omega": 0.2,
                        "pressure_crit": 200000.0,
                        "temperature_crit": 200,
                    },
                },
            },
            phases={
                "Vap": {
                    "type": VaporPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {
                        "type": CubicType.PR,
                        "mixing_rule_a": "wong_sandler",
                    },
                },
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                },
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            phases_in_equilibrium=[("Vap", "Liq")],
            phase_equilibrium_state={("Vap", "Liq"): modules[__name__]},
            parameter_data={
                "PR_kappa": {
                    ("a", "a"): 0.0,
                    ("a", "b"): 0.0,
                    ("b", "a"): 0.0,
                    ("b", "b"): 0.0,
                }
            },
        )


# This test can be resuscitated when we actually implement another mixing rule
# @pytest.mark.unit
# def test_mixing_rule_mismatch():
#     m = ConcreteModel()

#     # Dummy params block
#     with pytest.raises(ConfigurationError):
#         m.params = DummyParameterBlock(default={
#                     "components": {
#                         "a": {"phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
#                               "parameter_data": {
#                                   "omega": 0.1,
#                                   "pressure_crit": 1e5,
#                                   "temperature_crit": 100}},
#                         "b": {"phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
#                               "parameter_data": {
#                                   "omega": 0.2,
#                                   "pressure_crit": 2e5,
#                                   "temperature_crit": 200}}},
#                     "phases": {
#                         "Vap": {"type": VaporPhase,
#                                 "equation_of_state": Cubic,
#                                 "equation_of_state_options": {
#                                     "type": CubicType.PR,
#                                     "mixing_rule_a": MixingRuleA.default}},
#                         "Liq": {"type": LiquidPhase,
#                                 "equation_of_state": Cubic,
#                                 "equation_of_state_options": {
#                                     "type": CubicType.PR,
#                                     "mixing_rule_a": MixingRuleA.dummy}}},
#                     "base_units": {"time": pyunits.s,
#                                     "length": pyunits.m,
#                                     "mass": pyunits.kg,
#                                     "amount": pyunits.mol,
#                                     "temperature": pyunits.K},
#                     "state_definition": modules[__name__],
#                     "pressure_ref": 1e5,
#                     "temperature_ref": 300,
#                     "phases_in_equilibrium": [("Vap", "Liq")],
#                     "phase_equilibrium_state": {("Vap", "Liq"): modules[__name__]},
#                     "parameter_data": {"PR_kappa": {("a", "a"): 0.000,
#                                                     ("a", "b"): 0.000,
#                                                     ("b", "a"): 0.000,
#                                                     ("b", "b"): 0.000,}}})


@pytest.mark.unit
def test_config_fail():
    m = ConcreteModel()

    # Dummy params block
    with pytest.raises(ValueError):
        m.params = DummyParameterBlock(
            components={
                "a": {
                    "phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                    "parameter_data": {
                        "omega": 0.1,
                        "pressure_crit": 100000.0,
                        "temperature_crit": 100,
                    },
                },
                "b": {
                    "phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                    "parameter_data": {
                        "omega": 0.2,
                        "pressure_crit": 200000.0,
                        "temperature_crit": 200,
                    },
                },
            },
            phases={
                "Vap": {
                    "type": VaporPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {
                        "type": CubicType.PR,
                        "mixing_rule_a": "wong_sandler",
                    },
                },
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                },
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            phases_in_equilibrium=[("Vap", "Liq")],
            phase_equilibrium_state={("Vap", "Liq"): modules[__name__]},
            parameter_data={
                "PR_kappa": {
                    ("a", "a"): 0.0,
                    ("a", "b"): 0.0,
                    ("b", "a"): 0.0,
                    ("b", "b"): 0.0,
                }
            },
        )


@pytest.mark.unit
def test_common(m):
    # Test cubic components
    assert isinstance(m.props[1].PR_fw, Expression)
    assert len(m.props[1].PR_fw) == len(m.params.component_list)
    for i in m.params.component_list:
        omega = m.params.get_component(i).omega
        assert value(m.props[1].PR_fw[i]) == value(
            0.37464 + 1.54226 * omega - 0.26992 * omega**2
        )

    assert isinstance(m.props[1].PR_a, Expression)
    assert len(m.props[1].PR_a) == len(m.params.component_list)
    for i in m.params.component_list:
        Tc = m.params.get_component(i).temperature_crit
        Pc = m.params.get_component(i).pressure_crit
        assert pytest.approx(value(m.props[1].PR_a[i]), rel=1e-5) == value(
            0.45724
            * ((const.gas_constant * Tc) ** 2 / Pc)
            * (
                (
                    (1 + m.props[1].PR_fw[i] * (1 - sqrt(m.props[1].temperature / Tc)))
                    ** 2
                )
            )
        )

    assert isinstance(m.props[1].PR_b, Expression)
    assert len(m.props[1].PR_b) == len(m.params.component_list)
    for i in m.params.component_list:
        Tc = m.params.get_component(i).temperature_crit
        Pc = m.params.get_component(i).pressure_crit
        assert value(m.props[1].PR_b[i]) == value(
            0.07780 * const.gas_constant * Tc / Pc
        )

    assert isinstance(m.props[1].PR_am, Expression)
    assert len(m.props[1].PR_am) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_am[p]), rel=1e-5) == value(
            sum(
                sum(
                    m.props[1].mole_frac_phase_comp[p, i]
                    * m.props[1].mole_frac_phase_comp[p, j]
                    * sqrt(m.props[1].PR_a[i] * m.props[1].PR_a[j])
                    * (1 - m.params.PR_kappa[i, j])
                    for j in m.params.component_list
                )
                for i in m.params.component_list
            )
        )

    assert isinstance(m.props[1].PR_bm, Expression)
    assert len(m.props[1].PR_bm) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_bm[p]), rel=1e-5) == value(
            sum(
                m.props[1].mole_frac_phase_comp[p, i] * m.props[1].PR_b[i]
                for i in m.props[1].params.component_list
            )
        )

    assert isinstance(m.props[1].PR_A, Expression)
    assert len(m.props[1].PR_A) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_A[p]), rel=1e-5) == value(
            m.props[1].PR_am[p]
            * m.props[1].pressure
            / (const.gas_constant * m.props[1].temperature) ** 2
        )
    assert pytest.approx(value(m.props[1].PR_A["Liq"]), rel=1e-5) == Al
    assert pytest.approx(value(m.props[1].PR_A["Vap"]), rel=1e-5) == Av

    assert isinstance(m.props[1].PR_B, Expression)
    assert len(m.props[1].PR_B) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_B[p]), rel=1e-5) == value(
            m.props[1].PR_bm[p]
            * m.props[1].pressure
            / (const.gas_constant * m.props[1].temperature)
        )
    assert pytest.approx(value(m.props[1].PR_B["Liq"]), rel=1e-5) == Bl
    assert pytest.approx(value(m.props[1].PR_B["Vap"]), rel=1e-5) == Bv

    assert isinstance(m.props[1].PR_delta, Expression)
    assert len(m.props[1].PR_delta) == 6
    for p in m.params.phase_list:
        for i in m.params.component_list:
            assert pytest.approx(value(m.props[1].PR_delta[p, i]), rel=1e-5) == value(
                2
                * sqrt(m.props[1].PR_a[i])
                / m.props[1].PR_am[p]
                * sum(
                    m.props[1].mole_frac_phase_comp[p, j]
                    * sqrt(m.props[1].PR_a[j])
                    * (1 - m.params.PR_kappa[i, j])
                    for j in m.params.component_list
                )
            )

    assert isinstance(m.props[1].PR_daij_dT, Expression)
    assert len(m.props[1].PR_daij_dT) == len(m.params.component_list) ** 2
    for i in m.params.component_list:
        for j in m.params.component_list:
            assert value(m.props[1].PR_daij_dT[i, j]) == pytest.approx(
                value(
                    -(const.gas_constant / 2)
                    * sqrt(0.45724)
                    * (1 - m.params.PR_kappa[i, j])
                    / sqrt(m.props[1].temperature)
                    * (
                        m.props[1].PR_fw[j]
                        * sqrt(
                            m.props[1].PR_a[i]
                            * m.params.get_component(j).temperature_crit
                            / m.params.get_component(j).pressure_crit
                        )
                        + m.props[1].PR_fw[i]
                        * sqrt(
                            m.props[1].PR_a[j]
                            * m.params.get_component(i).temperature_crit
                            / m.params.get_component(i).pressure_crit
                        )
                    )
                )
            )

    assert isinstance(m.props[1].PR_dam_dT, Expression)
    assert len(m.props[1].PR_dam_dT) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_dam_dT[p]), rel=1e-5) == value(
            -(
                (const.gas_constant / 2)
                * sqrt(0.45724)
                * sum(
                    sum(
                        m.props[1].mole_frac_phase_comp[p, i]
                        * m.props[1].mole_frac_phase_comp[p, j]
                        * (1 - m.params.PR_kappa[i, j])
                        * (
                            m.props[1].PR_fw[j]
                            * sqrt(
                                m.props[1].PR_a[i]
                                * m.params.get_component(j).temperature_crit
                                / m.params.get_component(j).pressure_crit
                            )
                            + m.props[1].PR_fw[i]
                            * sqrt(
                                m.props[1].PR_a[j]
                                * m.params.get_component(i).temperature_crit
                                / m.params.get_component(i).pressure_crit
                            )
                        )
                        for j in m.params.component_list
                    )
                    for i in m.params.component_list
                )
                / sqrt(m.props[1].temperature)
            )
        )

    assert isinstance(m.props[1].PR_d2am_dT2, Expression)
    assert len(m.props[1].PR_d2am_dT2) == len(m.params.phase_list)
    T = m.props[1].temperature
    R = const.gas_constant
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_d2am_dT2[p])) == value(
            -(0.5 / T) * m.props[1].PR_dam_dT[p]
            + ((R**2 * 0.45724) / (2 * T))
            * sum(
                sum(
                    m.props[1].mole_frac_phase_comp[p, i]
                    * m.props[1].mole_frac_phase_comp[p, j]
                    * (1 - m.params.PR_kappa[i, j])
                    * m.props[1].PR_fw[i]
                    * m.props[1].PR_fw[j]
                    * sqrt(
                        (
                            m.params.get_component(i).temperature_crit
                            * m.params.get_component(j).temperature_crit
                        )
                        / (
                            m.params.get_component(i).pressure_crit
                            * m.params.get_component(j).pressure_crit
                        )
                    )
                    for j in m.params.component_list
                )
                for i in m.params.component_list
            )
        )

    # Test equilibrium state Expressions
    assert isinstance(m.props[1]._PR_a_eq, Expression)
    assert len(m.props[1]._PR_a_eq) == len(m.params.component_list)
    for i in m.props[1]._PR_a_eq:
        Tc = m.params.get_component(i[2]).temperature_crit
        Pc = m.params.get_component(i[2]).pressure_crit
        assert pytest.approx(value(m.props[1]._PR_a_eq[i]), rel=1e-5) == value(
            0.45724
            * ((const.gas_constant * Tc) ** 2 / Pc)
            * (
                (
                    (
                        1
                        + m.props[1].PR_fw[i[2]]
                        * (1 - sqrt(m.props[1]._teq[i[0], i[1]] / Tc))
                    )
                    ** 2
                )
            )
        )

    assert isinstance(m.props[1]._PR_am_eq, Expression)
    assert len(m.props[1]._PR_am_eq) == len(m.params.phase_list)
    for idx in m.props[1]._PR_am_eq:
        assert pytest.approx(value(m.props[1]._PR_am_eq[idx]), rel=1e-5) == value(
            sum(
                sum(
                    m.props[1].mole_frac_phase_comp[idx[2], i]
                    * m.props[1].mole_frac_phase_comp[idx[2], j]
                    * sqrt(
                        m.props[1]._PR_a_eq[idx[0], idx[1], i]
                        * m.props[1]._PR_a_eq[idx[0], idx[1], j]
                    )
                    * (1 - m.params.PR_kappa[i, j])
                    for j in m.params.component_list
                )
                for i in m.params.component_list
            )
        )

    assert isinstance(m.props[1]._PR_A_eq, Expression)
    assert len(m.props[1]._PR_A_eq) == len(m.params.phase_list)
    for i in m.props[1]._PR_A_eq:
        assert pytest.approx(value(m.props[1]._PR_A_eq[i]), rel=1e-5) == value(
            m.props[1]._PR_am_eq[i]
            * m.props[1].pressure
            / (const.gas_constant * m.props[1]._teq[i[0], i[1]]) ** 2
        )
    assert (
        pytest.approx(value(m.props[1]._PR_A_eq[("Vap", "Liq"), "Liq"]), rel=1e-5)
        == Al_eq
    )
    assert (
        pytest.approx(value(m.props[1]._PR_A_eq[("Vap", "Liq"), "Vap"]), rel=1e-5)
        == Av_eq
    )

    assert isinstance(m.props[1]._PR_B_eq, Expression)
    assert len(m.props[1]._PR_B_eq) == len(m.params.phase_list)
    for i in m.props[1]._PR_B_eq:
        assert pytest.approx(value(m.props[1]._PR_B_eq[i]), rel=1e-5) == value(
            m.props[1].PR_bm[i[2]]
            * m.props[1].pressure
            / (const.gas_constant * m.props[1]._teq[i[0], i[1]])
        )
    assert (
        pytest.approx(value(m.props[1]._PR_B_eq[("Vap", "Liq"), "Liq"]), rel=1e-5)
        == Bl_eq
    )
    assert (
        pytest.approx(value(m.props[1]._PR_B_eq[("Vap", "Liq"), "Vap"]), rel=1e-5)
        == Bv_eq
    )

    assert isinstance(m.props[1]._PR_delta_eq, Expression)
    assert len(m.props[1]._PR_delta_eq) == 6
    for idx in m.props[1]._PR_delta_eq:
        assert pytest.approx(value(m.props[1]._PR_delta_eq[idx]), rel=1e-5) == value(
            2
            * sqrt(m.props[1]._PR_a_eq[idx[0], idx[1], idx[3]])
            / m.props[1]._PR_am_eq[idx[0], idx[1], idx[2]]
            * sum(
                m.props[1].mole_frac_phase_comp[idx[2], j]
                * sqrt(m.props[1]._PR_a_eq[idx[0], idx[1], j])
                * (1 - m.params.PR_kappa[idx[3], j])
                for j in m.params.component_list
            )
        )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_compress_fact_phase_Liq(m):
    assert isinstance(
        Cubic.compress_fact_phase(m.props[1], "Liq"), ExternalFunctionExpression
    )
    assert (
        pytest.approx(value(Cubic.compress_fact_phase(m.props[1], "Liq")), rel=1e-5)
        == Zl
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_compress_fact_phase_Vap(m):
    assert isinstance(
        Cubic.compress_fact_phase(m.props[1], "Vap"), ExternalFunctionExpression
    )
    assert (
        pytest.approx(value(Cubic.compress_fact_phase(m.props[1], "Vap")), rel=1e-5)
        == Zv
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_dens_mass_phase(m):
    for p in m.params.phase_list:
        assert str(Cubic.dens_mass_phase(m.props[1], p)) == str(
            m.props[1].dens_mol_phase[p] * m.props[1].mw_phase[p]
        )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_dens_mol_phase(m):
    assert value(Cubic.dens_mol_phase(m.props[1], "Vap")) == pytest.approx(
        44.800, rel=1e-3
    )
    assert value(Cubic.dens_mol_phase(m.props[1], "Liq")) == pytest.approx(
        41.157, rel=1e-3
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_energy_internal_mol_phase(m):
    assert pytest.approx(Uv, rel=1e-4) == value(
        Cubic.energy_internal_mol_phase(m.props[1], "Vap")
    )
    assert pytest.approx(Ul, rel=1e-4) == value(
        Cubic.energy_internal_mol_phase(m.props[1], "Liq")
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_energy_internal_mol_phase_comp(m):
    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert str(Cubic.energy_internal_mol_phase_comp(m.props[1], p, j)) == str(
                m.props[1].enth_mol_phase_comp[p, j]
                - m.props[1].pressure * m.props[1].vol_mol_phase_comp[p, j]
            )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_enth_mol_phase(m):
    assert pytest.approx(value(Cubic.enth_mol_phase(m.props[1], "Vap")), rel=1e-5) == Hv
    assert pytest.approx(value(Cubic.enth_mol_phase(m.props[1], "Liq")), rel=1e-5) == Hl


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_enth_mol_phase_comp(m):
    enth = {
        ("Liq", "a"): -23.8270,
        ("Vap", "a"): 345.7288,
        ("Liq", "b"): -407.3709,
        ("Vap", "b"): -385.7457,
        ("Liq", "c"): -757.0346,
        ("Vap", "c"): -1052.4697,
    }
    for j in m.params.component_list:
        assert pytest.approx(enth["Liq", j], rel=1e-5) == value(
            Cubic.enth_mol_phase_comp(m.props[1], "Liq", j)
        )
        assert pytest.approx(enth["Vap", j], rel=1e-5) == value(
            Cubic.enth_mol_phase_comp(m.props[1], "Vap", j)
        )

    assert pytest.approx(Hl, rel=1e-5) == value(
        sum(
            Cubic.enth_mol_phase_comp(m.props[1], "Liq", j)
            * m.props[1].mole_frac_phase_comp["Liq", j]
            for j in m.params.component_list
        )
    )
    assert pytest.approx(Hv, rel=1e-5) == value(
        sum(
            Cubic.enth_mol_phase_comp(m.props[1], "Vap", j)
            * m.props[1].mole_frac_phase_comp["Vap", j]
            for j in m.params.component_list
        )
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_entr_mol_phase(m):
    assert (
        pytest.approx(value(Cubic.entr_mol_phase(m.props[1], "Vap")), rel=1e-5)
        == 46.58858
    )
    assert (
        pytest.approx(value(Cubic.entr_mol_phase(m.props[1], "Liq")), rel=1e-5)
        == 48.62807
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_entr_mol_phase_comp(m):
    entr = {
        ("Liq", "a"): 45.8181,
        ("Vap", "a"): 61.5670,
        ("Liq", "b"): 50.7494,
        ("Vap", "b"): 54.1797,
        ("Liq", "c"): 59.1236,
        ("Vap", "c"): 42.2799,
    }
    for j in m.params.component_list:
        assert pytest.approx(entr[("Liq", j)], rel=1e-5) == value(
            Cubic.entr_mol_phase_comp(m.props[1], "Liq", j)
        )
        assert pytest.approx(entr[("Vap", j)], rel=1e-5) == value(
            Cubic.entr_mol_phase_comp(m.props[1], "Vap", j)
        )

    # With real partial molar quantities implemented, we should have that
    # the sum weighted by mole fraction is equal to the molar entropy
    assert pytest.approx(
        value(Cubic.entr_mol_phase(m.props[1], "Liq")), rel=1e-5
    ) == value(
        sum(
            Cubic.entr_mol_phase_comp(m.props[1], "Liq", j)
            * m.props[1].mole_frac_phase_comp["Liq", j]
            for j in m.params.component_list
        )
    )
    assert pytest.approx(
        value(Cubic.entr_mol_phase(m.props[1], "Vap")), rel=1e-5
    ) == value(
        sum(
            Cubic.entr_mol_phase_comp(m.props[1], "Vap", j)
            * m.props[1].mole_frac_phase_comp["Vap", j]
            for j in m.params.component_list
        )
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.component
def test_fug_phase_comp(m):
    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert str(Cubic.fug_phase_comp(m.props[1], p, j)) == str(
                m.props[1].mole_frac_phase_comp[p, j]
                * m.props[1].pressure
                * m.props[1].fug_coeff_phase_comp[p, j]
            )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.component
def test_fug_phase_comp_eq(m):
    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert str(
                Cubic.fug_phase_comp_eq(m.props[1], p, j, ("Vap", "Liq"))
            ) == str(
                m.props[1].mole_frac_phase_comp[p, j]
                * m.props[1].pressure
                * Cubic.fug_coeff_phase_comp_eq(m.props[1], p, j, ("Vap", "Liq"))
            )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_fug_coeff_phase_comp_Liq(m):
    assert pytest.approx(1.01213, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Liq", "a")
    )
    assert pytest.approx(0.95919, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Liq", "b")
    )
    assert pytest.approx(0.91356, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Liq", "c")
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_fug_coeff_phase_comp_Vap(m):
    assert pytest.approx(1.05952, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Vap", "a")
    )
    assert pytest.approx(0.96070, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Vap", "b")
    )
    assert pytest.approx(0.87903, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Vap", "c")
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_fug_coeff_phase_comp_eq_Liq(m):
    assert pytest.approx(1.22431, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Liq", "a", ("Vap", "Liq"))
    )
    assert pytest.approx(0.0049604, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Liq", "b", ("Vap", "Liq"))
    )
    assert pytest.approx(3.19128e-05, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Liq", "c", ("Vap", "Liq"))
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_fug_coeff_phase_comp_eq_Vap(m):
    assert pytest.approx(86.9140, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Vap", "a", ("Vap", "Liq"))
    )
    assert pytest.approx(0.0049878, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Vap", "b", ("Vap", "Liq"))
    )
    assert pytest.approx(6.49776e-07, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Vap", "c", ("Vap", "Liq"))
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_gibbs_mol_phase(m):
    for p in m.params.phase_list:
        assert str(Cubic.gibbs_mol_phase(m.props[1], p)) == str(
            m.props[1].enth_mol_phase[p]
            - m.props[1].entr_mol_phase[p] * m.props[1].temperature
        )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_gibbs_mol_phase_comp(m):
    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert pytest.approx(
                value(Cubic.gibbs_mol_phase_comp(m.props[1], p, j)), rel=1e-5
            ) == value(
                Cubic.enth_mol_phase_comp(m.props[1], p, j)
                - Cubic.entr_mol_phase_comp(m.props[1], p, j) * m.props[1].temperature
            )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_vol_mol_phase(m):
    assert value(Cubic.vol_mol_phase(m.props[1], "Vap")) == pytest.approx(
        1 / 44.800, rel=1e-3
    )
    assert value(Cubic.vol_mol_phase(m.props[1], "Liq")) == pytest.approx(
        1 / 41.157, rel=1e-3
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_vol_mol_phase_comp(m):
    vol = {
        ("Liq", "a"): 0.0249368,
        ("Vap", "a"): 0.0262974,
        ("Liq", "b"): 0.0236340,
        ("Vap", "b"): 0.0237075,
        ("Liq", "c"): 0.0224518,
        ("Vap", "c"): 0.0213573,
    }
    for j in m.params.component_list:
        assert pytest.approx(vol[("Liq", j)], rel=1e-5) == value(
            Cubic.vol_mol_phase_comp(m.props[1], "Liq", j)
        )
        assert pytest.approx(vol[("Vap", j)], rel=1e-5) == value(
            Cubic.vol_mol_phase_comp(m.props[1], "Vap", j)
        )

    # With real partial molar quantities implemented, we should have that
    # the sum weighted by mole fraction is equal to the molar volume
    assert pytest.approx(
        value(Cubic.vol_mol_phase(m.props[1], "Liq")), rel=1e-5
    ) == value(
        sum(
            Cubic.vol_mol_phase_comp(m.props[1], "Liq", j)
            * m.props[1].mole_frac_phase_comp["Liq", j]
            for j in m.params.component_list
        )
    )
    assert pytest.approx(
        value(Cubic.vol_mol_phase(m.props[1], "Vap")), rel=1e-5
    ) == value(
        sum(
            Cubic.vol_mol_phase_comp(m.props[1], "Vap", j)
            * m.props[1].mole_frac_phase_comp["Vap", j]
            for j in m.params.component_list
        )
    )


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
@pytest.mark.unit
def test_calculate_equilibrium_cubic_coefficients(m):
    b, c, d = calculate_equilibrium_cubic_coefficients(
        m.props[1], "PR", CubicType.PR, "Vap", "Liq", "Liq"
    )

    A_eq = m.props[1]._PR_A_eq["Vap", "Liq", "Liq"]
    B_eq = m.props[1]._PR_B_eq["Vap", "Liq", "Liq"]
    EoS_u = EoS_param[CubicType.PR]["u"]
    EoS_w = EoS_param[CubicType.PR]["w"]

    assert str(b) == str(-(1 + B_eq - EoS_u * B_eq))
    assert str(c) == str(A_eq - EoS_u * B_eq - EoS_u * B_eq**2 + EoS_w * B_eq**2)
    assert str(d) == str(-(A_eq * B_eq + EoS_w * B_eq**2 + EoS_w * B_eq**3))


class TestCEOSCriticalProps:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        # Dummy params block
        m.params = GenericParameterBlock(
            components={
                "a": {
                    "parameter_data": {
                        "omega": 0.1,
                        "compress_fact_crit": 1,
                        "dens_mol_crit": 1,
                        "pressure_crit": 1e5,
                        "temperature_crit": 100,
                    }
                },
                "b": {
                    "parameter_data": {
                        "omega": 0.2,
                        "compress_fact_crit": 1,
                        "dens_mol_crit": 1,
                        "pressure_crit": 2e5,
                        "temperature_crit": 200,
                    }
                },
                "c": {
                    "parameter_data": {
                        "omega": 0.3,
                        "compress_fact_crit": 1,
                        "dens_mol_crit": 1,
                        "pressure_crit": 3e5,
                        "temperature_crit": 300,
                    }
                },
            },
            phases={
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                },
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_definition=FTPx,
            pressure_ref=100000.0,
            temperature_ref=300,
            parameter_data={
                "PR_kappa": {
                    ("a", "a"): 0.0,
                    ("a", "b"): 0.0,
                    ("a", "c"): 0.0,
                    ("b", "a"): 0.0,
                    ("b", "b"): 0.0,
                    ("b", "c"): 0.0,
                    ("c", "a"): 0.0,
                    ("c", "b"): 0.0,
                    ("c", "c"): 0.0,
                },
            },
        )

        m.props = m.params.build_state_block(
            [1], defined_state=True, parameters=m.params
        )

        # Fix state variables
        m.props[1].flow_mol.fix(1)
        m.props[1].pressure.fix(1e5)
        m.props[1].temperature.fix(300)

        m.props[1].mole_frac_comp["a"].fix(0.4)
        m.props[1].mole_frac_comp["b"].fix(0.6)
        m.props[1].mole_frac_comp["c"].fix(1e-8)

        # Build critical properties
        m.props[1]._critical_props()

        return m

    @pytest.mark.unit
    def test_build_critical_props(self, model):
        assert isinstance(model.props[1].a_crit, Expression)
        assert isinstance(model.props[1].am_crit, Expression)
        assert isinstance(model.props[1].bm_crit, Expression)

        base_units = model.params.get_metadata().default_units

        assert str(model.props[1].a_crit["a"].expr) == str(
            0.45724
            * (
                (
                    pyunits.convert(
                        const.gas_constant, to_units=base_units.GAS_CONSTANT
                    )
                    * model.params.a.temperature_crit
                )
                ** 2
                / model.params.a.pressure_crit
            )
            * (
                1
                + model.props[1].PR_fw["a"]
                * (
                    1
                    - sqrt(
                        model.props[1].temperature_crit
                        / model.params.a.temperature_crit
                    )
                )
            )
            ** 2
        )
        assert str(model.props[1].a_crit["b"].expr) == str(
            0.45724
            * (
                (
                    pyunits.convert(
                        const.gas_constant, to_units=base_units.GAS_CONSTANT
                    )
                    * model.params.b.temperature_crit
                )
                ** 2
                / model.params.b.pressure_crit
            )
            * (
                1
                + model.props[1].PR_fw["b"]
                * (
                    1
                    - sqrt(
                        model.props[1].temperature_crit
                        / model.params.b.temperature_crit
                    )
                )
            )
            ** 2
        )
        assert str(model.props[1].a_crit["c"].expr) == str(
            0.45724
            * (
                (
                    pyunits.convert(
                        const.gas_constant, to_units=base_units.GAS_CONSTANT
                    )
                    * model.params.c.temperature_crit
                )
                ** 2
                / model.params.c.pressure_crit
            )
            * (
                1
                + model.props[1].PR_fw["c"]
                * (
                    1
                    - sqrt(
                        model.props[1].temperature_crit
                        / model.params.c.temperature_crit
                    )
                )
            )
            ** 2
        )

        assert str(model.props[1].am_crit.expr) == str(
            (
                model.props[1].mole_frac_comp["a"]
                * model.props[1].mole_frac_comp["a"]
                * sqrt(model.props[1].a_crit["a"] * model.props[1].a_crit["a"])
                * (1 - model.params.PR_kappa["a", "a"])
            )
            + (
                model.props[1].mole_frac_comp["a"]
                * model.props[1].mole_frac_comp["b"]
                * sqrt(model.props[1].a_crit["a"] * model.props[1].a_crit["b"])
                * (1 - model.params.PR_kappa["a", "b"])
            )
            + (
                model.props[1].mole_frac_comp["a"]
                * model.props[1].mole_frac_comp["c"]
                * sqrt(model.props[1].a_crit["a"] * model.props[1].a_crit["c"])
                * (1 - model.params.PR_kappa["a", "c"])
            )
            + (
                model.props[1].mole_frac_comp["b"]
                * model.props[1].mole_frac_comp["a"]
                * sqrt(model.props[1].a_crit["b"] * model.props[1].a_crit["a"])
                * (1 - model.params.PR_kappa["b", "a"])
            )
            + (
                model.props[1].mole_frac_comp["b"]
                * model.props[1].mole_frac_comp["b"]
                * sqrt(model.props[1].a_crit["b"] * model.props[1].a_crit["b"])
                * (1 - model.params.PR_kappa["b", "b"])
            )
            + (
                model.props[1].mole_frac_comp["b"]
                * model.props[1].mole_frac_comp["c"]
                * sqrt(model.props[1].a_crit["b"] * model.props[1].a_crit["c"])
                * (1 - model.params.PR_kappa["b", "c"])
            )
            + (
                model.props[1].mole_frac_comp["c"]
                * model.props[1].mole_frac_comp["a"]
                * sqrt(model.props[1].a_crit["c"] * model.props[1].a_crit["a"])
                * (1 - model.params.PR_kappa["c", "a"])
            )
            + (
                model.props[1].mole_frac_comp["c"]
                * model.props[1].mole_frac_comp["b"]
                * sqrt(model.props[1].a_crit["c"] * model.props[1].a_crit["b"])
                * (1 - model.params.PR_kappa["c", "b"])
            )
            + (
                model.props[1].mole_frac_comp["c"]
                * model.props[1].mole_frac_comp["c"]
                * sqrt(model.props[1].a_crit["c"] * model.props[1].a_crit["c"])
                * (1 - model.params.PR_kappa["c", "c"])
            )
        )

        assert str(model.props[1].bm_crit.expr) == str(
            model.props[1].mole_frac_comp["a"] * model.props[1].PR_b["a"]
            + model.props[1].mole_frac_comp["b"] * model.props[1].PR_b["b"]
            + model.props[1].mole_frac_comp["c"] * model.props[1].PR_b["c"]
        )

        assert isinstance(model.props[1].A_crit, Constraint)
        assert isinstance(model.props[1].B_crit, Constraint)

        assert str(model.props[1].A_crit.expr) == str(
            0.45724
            * (
                pyunits.convert(const.gas_constant, to_units=base_units.GAS_CONSTANT)
                * model.props[1].temperature_crit
            )
            ** 2
            == model.props[1].am_crit * model.props[1].pressure_crit
        )

        assert str(model.props[1].B_crit.expr) == str(
            0.07780
            * pyunits.convert(const.gas_constant, to_units=base_units.GAS_CONSTANT)
            * model.props[1].temperature_crit
            == model.props[1].bm_crit * model.props[1].pressure_crit
        )

        assert_units_consistent(model)

        # Check list_critical_property_constraint_names
        con_list = Cubic.list_critical_property_constraint_names()

        assert con_list == [
            "dens_mol_crit_eq",
            "compress_fact_crit_eq",
            "A_crit",
            "B_crit",
        ]

        # Check that names match constraints that were built
        for c in con_list:
            assert isinstance(getattr(model.props[1], c), Constraint)

    @pytest.mark.component
    @pytest.mark.solver
    def test_initialize_with_critical_props(self, model):
        initializer = ModularPropertiesInitializer()

        status = initializer.initialize(model.props)

        assert status == InitializationStatus.Ok

        # Confirm results
        assert value(model.props[1].compress_fact_crit) == pytest.approx(
            0.321379, rel=1e-5
        )
        assert value(model.props[1].pressure_crit) == pytest.approx(1.58138e5, rel=1e-5)
        assert value(model.props[1].temperature_crit) == pytest.approx(
            158.138, rel=1e-5
        )
        assert value(model.props[1].dens_mol_crit) == pytest.approx(374.238, rel=1e-5)

    @pytest.mark.component
    @pytest.mark.solver
    def test_solve_critical_props(self, model):
        # Solve model
        solver = get_solver("ipopt_v2")
        res = solver.solve(model, tee=True)
        assert_optimal_termination(res)

        # Confirm results
        assert value(model.props[1].compress_fact_crit) == pytest.approx(
            0.321379, rel=1e-5
        )
        assert value(model.props[1].pressure_crit) == pytest.approx(1.58138e5, rel=1e-5)
        assert value(model.props[1].temperature_crit) == pytest.approx(
            158.138, rel=1e-5
        )
        assert value(model.props[1].dens_mol_crit) == pytest.approx(374.238, rel=1e-5)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Framework for generic reaction packages
"""
# TODO: Look into protected access issues
# pylint: disable=protected-access

from math import log

# Import Pyomo libraries
from pyomo.environ import Block, Constraint, Expression, Set, units as pyunits, Var
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    ReactionParameterBlock,
    ReactionBlockDataBase,
    ReactionBlockBase,
    MaterialFlowBasis,
    ElectrolytePropertySet,
)
from idaes.models.properties.modular_properties.base.utility import ConcentrationForm
from idaes.core.util.exceptions import (
    BurntToast,
    ConfigurationError,
    PropertyPackageError,
)
import idaes.core.util.scaling as iscale

import idaes.logger as idaeslog


# Set up logger
_log = idaeslog.getLogger(__name__)


class GenericReactionPackageError(PropertyPackageError):
    """Modular reaction package class."""

    # Error message for when a property is called for but no option provided
    def __init__(self, block, prop):
        super().__init__()
        self.prop = prop
        self.block = block

    def __str__(self):
        return (
            f"Generic Reaction Package instance {self.block} called for "
            f"{self.prop}, but was not provided with a method "
            f"for this property. Please add a method for this property "
            f"in the reaction parameter configuration."
        )


rxn_config = ConfigBlock()
rxn_config.declare(
    "stoichiometry",
    ConfigValue(
        domain=dict,
        description="Stoichiometry of reaction",
        doc="Dict describing stoichiometry of reaction",
    ),
)
rxn_config.declare(
    "heat_of_reaction",
    ConfigValue(
        description="Method for calculating specific heat of reaction",
        doc="Valid Python class containing instructions on how to calculate "
        "the heat of reaction for this reaction.",
    ),
)
rxn_config.declare(
    "concentration_form",
    ConfigValue(
        default=None,
        domain=In(ConcentrationForm),
        description="Form to use for concentration terms in reaction equation",
        doc="ConcentrationForm Enum indicating what form to use for concentration "
        "terms when constructing reaction equation.",
    ),
)
rxn_config.declare(
    "parameter_data",
    ConfigValue(
        default={},
        domain=dict,
        description="Dict containing initialization data for parameters",
    ),
)

rate_rxn_config = rxn_config()
rate_rxn_config.declare(
    "rate_constant",
    ConfigValue(
        description="Expression form describing rate constant",
        doc="Valid Python class containing instructions on how to construct "
        "the rate constant for this reaction.",
    ),
)
rate_rxn_config.declare(
    "rate_form",
    ConfigValue(
        description="Expression form describing rate of reaction",
        doc="Valid Python class containing instructions on how to construct "
        "the rate expression for this reaction.",
    ),
)

equil_rxn_config = rxn_config()
equil_rxn_config.declare(
    "equilibrium_constant",
    ConfigValue(
        description="Expression form describing equilibrium constant",
        doc="Valid Python class containing instructions on how to construct "
        "the equilibrium constant for this reaction.",
    ),
)
equil_rxn_config.declare(
    "equilibrium_form",
    ConfigValue(
        description="Expression form describing reaction equilibrium",
        doc="Valid Python class containing instructions on how to construct "
        "the equilibrium constraint for this reaction.",
    ),
)


@declare_process_block_class("GenericReactionParameterBlock")
class GenericReactionParameterData(ReactionParameterBlock):
    """
    General Reaction Parameter Block Class
    """

    CONFIG = ReactionParameterBlock.CONFIG()

    CONFIG.declare(
        "reaction_basis",
        ConfigValue(
            default=MaterialFlowBasis.molar,
            domain=In(MaterialFlowBasis),
            doc="Basis of reactions",
            description="Argument indicating basis of reaction terms. Should be "
            "an instance of a MaterialFlowBasis Enum",
        ),
    )

    CONFIG.declare(
        "rate_reactions", ConfigBlock(implicit=True, implicit_domain=rate_rxn_config)
    )

    CONFIG.declare(
        "equilibrium_reactions",
        ConfigBlock(implicit=True, implicit_domain=equil_rxn_config),
    )

    # Base units of measurement
    CONFIG.declare(
        "base_units",
        ConfigValue(
            default={},
            domain=dict,
            description="Base units for property package",
            doc="Dict containing definition of base units of measurement to use "
            "with property package.",
        ),
    )

    # User-defined default scaling factors
    CONFIG.declare(
        "default_scaling_factors",
        ConfigValue(
            domain=dict,
            description="User-defined default scaling factors",
            doc="Dict of user-defined properties and associated default "
            "scaling factors",
        ),
    )

    def build(self):
        """
        Callable method for Block construction.
        """
        # Call super.build() to initialize Block
        # In this case we are replicating the super.build to get around a
        # chicken-and-egg problem
        # The super.build tries to validate units, but they have not been set
        # and cannot be set until the config block is created by super.build
        super(ReactionParameterBlock, self).build()

        # Check to make sure a property block was assigned
        if self.config.property_package is None:
            raise ConfigurationError(
                f"{self.name} was not assigned a property package. "
                "Did you forget to unpack the configuration dictionary?"
            )

        # Set base units of measurement
        self.get_metadata().add_default_units(self.config.base_units)

        # TODO: Need way to tie reaction package to a specific property package
        self._validate_property_parameter_units()
        self._validate_property_parameter_properties()

        # Call configure method to set construction arguments
        self.configure()

        # Build core components
        self._reaction_block_class = GenericReactionBlock

        # Alias associated property package to keep line length down
        ppack = self.config.property_package

        if not hasattr(ppack, "_electrolyte") or not ppack._electrolyte:
            pc_set = ppack.get_phase_component_set()
        elif ppack.config.state_components.name == "true":
            pc_set = ppack.true_phase_component_set
        elif ppack.config.state_components.name == "apparent":
            pc_set = ppack.apparent_phase_component_set
        else:
            raise BurntToast()

        # Construct rate reaction attributes if required
        if len(self.config.rate_reactions) > 0:
            # Construct rate reaction index
            self.rate_reaction_idx = Set(initialize=self.config.rate_reactions.keys())

            # Construct rate reaction stoichiometry dict
            self.rate_reaction_stoichiometry = {}
            for r, rxn in self.config.rate_reactions.items():
                for p, j in pc_set:
                    self.rate_reaction_stoichiometry[(r, p, j)] = 0

                if rxn.stoichiometry is None:
                    raise ConfigurationError(
                        "{} rate reaction {} was not provided with a "
                        "stoichiometry configuration argument.".format(self.name, r)
                    )
                else:
                    for k, v in rxn.stoichiometry.items():
                        if k[0] not in ppack.phase_list:
                            raise ConfigurationError(
                                "{} stoichiometry for rate reaction {} "
                                "included unrecognised phase {}.".format(
                                    self.name, r, k[0]
                                )
                            )
                        if k[1] not in ppack.component_list:
                            raise ConfigurationError(
                                "{} stoichiometry for rate reaction {} "
                                "included unrecognised component {}.".format(
                                    self.name, r, k[1]
                                )
                            )
                        self.rate_reaction_stoichiometry[(r, k[0], k[1])] = v

                # Check that a method was provided for the rate form
                if rxn.rate_form is None:
                    _log.debug(
                        "{} rate reaction {} was not provided with a "
                        "rate_form configuration argument. This is suitable "
                        "for processes using stoichiometric reactors, but not "
                        "for those using unit operations which rely on "
                        "reaction rate.".format(self.name, r)
                    )

        # Construct equilibrium reaction attributes if required
        if len(self.config.equilibrium_reactions) > 0:
            # Construct equilibrium reaction index
            self.equilibrium_reaction_idx = Set(
                initialize=self.config.equilibrium_reactions.keys()
            )

            # Construct equilibrium reaction stoichiometry dict
            self.equilibrium_reaction_stoichiometry = {}
            for r, rxn in self.config.equilibrium_reactions.items():
                for p, j in pc_set:
                    self.equilibrium_reaction_stoichiometry[(r, p, j)] = 0

                if rxn.stoichiometry is None:
                    raise ConfigurationError(
                        "{} equilibrium reaction {} was not provided with a "
                        "stoichiometry configuration argument.".format(self.name, r)
                    )
                else:
                    for k, v in rxn.stoichiometry.items():
                        if k[0] not in ppack.phase_list:
                            raise ConfigurationError(
                                "{} stoichiometry for equilibrium reaction {} "
                                "included unrecognised phase {}.".format(
                                    self.name, r, k[0]
                                )
                            )
                        if k[1] not in ppack.component_list:
                            raise ConfigurationError(
                                "{} stoichiometry for equilibrium reaction {} "
                                "included unrecognised component {}.".format(
                                    self.name, r, k[1]
                                )
                            )
                        self.equilibrium_reaction_stoichiometry[(r, k[0], k[1])] = v

                # Check that a method was provided for the equilibrium form
                if rxn.equilibrium_form is None:
                    raise ConfigurationError(
                        "{} equilibrium reaction {} was not provided with a "
                        "equilibrium_form configuration argument.".format(self.name, r)
                    )

        # Add a master reaction index which includes both types of reactions
        if (
            len(self.config.rate_reactions) > 0
            and len(self.config.equilibrium_reactions) > 0
        ):
            self.reaction_idx = Set(
                initialize=(self.rate_reaction_idx | self.equilibrium_reaction_idx)
            )
        elif len(self.config.rate_reactions) > 0:
            self.reaction_idx = Set(initialize=self.rate_reaction_idx)
        elif len(self.config.equilibrium_reactions) > 0:
            self.reaction_idx = Set(initialize=self.equilibrium_reaction_idx)
        else:
            raise BurntToast(
                "{} Generic property package failed to construct "
                "master reaction Set. This should not happen. "
                "Please contact the IDAES developers with this "
                "bug".format(self.name)
            )

        # Construct blocks to contain parameters for each reaction
        for r in self.reaction_idx:
            self.add_component("reaction_" + str(r), Block())

        # Build parameters
        if len(self.config.rate_reactions) > 0:
            for r in self.rate_reaction_idx:
                rblock = getattr(self, "reaction_" + r)
                r_config = self.config.rate_reactions[r]

                order_init = {}
                for p, j in pc_set:
                    if "reaction_order" in r_config.parameter_data:
                        try:
                            order_init[p, j] = r_config.parameter_data[
                                "reaction_order"
                            ][p, j]
                        except KeyError:
                            order_init[p, j] = 0
                    else:
                        # Assume elementary reaction and use stoichiometry
                        try:
                            if r_config.stoichiometry[p, j] < 0:
                                # These are reactants, but order is -ve stoic
                                order_init[p, j] = -r_config.stoichiometry[p, j]
                            else:
                                # Anything else is a product, not be included
                                order_init[p, j] = 0
                        except KeyError:
                            order_init[p, j] = 0

                rblock.reaction_order = Var(
                    pc_set, initialize=order_init, doc="Reaction order", units=None
                )

                for val in self.config.rate_reactions[r].values():
                    try:
                        val.build_parameters(rblock, self.config.rate_reactions[r])
                    except AttributeError:
                        pass

        if len(self.config.equilibrium_reactions) > 0:
            for r in self.equilibrium_reaction_idx:
                rblock = getattr(self, "reaction_" + r)
                r_config = self.config.equilibrium_reactions[r]

                order_init = {}
                for p, j in pc_set:
                    if "reaction_order" in r_config.parameter_data:
                        try:
                            order_init[p, j] = r_config.parameter_data[
                                "reaction_order"
                            ][p, j]
                        except KeyError:
                            order_init[p, j] = 0
                    else:
                        # Assume elementary reaction and use stoichiometry
                        try:
                            # Here we use the stoic. coeff. directly
                            # However, solids should be excluded as they
                            # normally do not appear in the equilibrium
                            # relationship
                            pobj = ppack.get_phase(p)
                            if not pobj.is_solid_phase():
                                order_init[p, j] = r_config.stoichiometry[p, j]
                            else:
                                order_init[p, j] = 0
                        except KeyError:
                            order_init[p, j] = 0

                rblock.reaction_order = Var(
                    pc_set, initialize=order_init, doc="Reaction order", units=None
                )

                for val in self.config.equilibrium_reactions[r].values():
                    try:
                        val.build_parameters(
                            rblock, self.config.equilibrium_reactions[r]
                        )
                    except AttributeError:
                        pass
                    except KeyError as err:
                        # This likely arises from mismatched true and apparent
                        # species sets. Reaction packages must use the same
                        # basis as the associated thermo properties
                        # Raise an exception to inform the user
                        raise PropertyPackageError(
                            "{} KeyError encountered whilst constructing "
                            "reaction parameters. This may be due to "
                            "mismatched state_components between the "
                            "Reaction Package and the associated Physical "
                            "Property Package - Reaction Packages must use the"
                            "same basis (true or apparent species) as the "
                            "Physical Property Package.".format(self.name),
                            err,
                        )

        # As a safety check, make sure all Vars in reaction blocks are fixed
        for v in self.component_objects(Var, descend_into=True):
            for i in v:
                if v[i].value is None:
                    raise ConfigurationError(
                        "{} parameter {} was not assigned"
                        " a value. Please check your configuration "
                        "arguments.".format(self.name, v.local_name)
                    )
                v[i].fix()

        # Set default scaling factors
        if self.config.default_scaling_factors is not None:
            self.default_scaling_factor.update(self.config.default_scaling_factors)
        # Finally, call populate_default_scaling_factors method to fill blanks
        iscale.populate_default_scaling_factors(self)

    def configure(self):
        """
        Placeholder method to allow users to specify config arguments via a
        class. The user class should inherit from this one and implement a
        configure() method which sets the values of the desired config
        arguments.

        Args:
            None

        Returns:
            None
        """

    def parameters(self):
        """
        Placeholder method to allow users to specify parameters via a
        class. The user class should inherit from this one and implement a
        parameters() method which creates the required components.

        Args:
            None

        Returns:
            None
        """

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.define_property_set(ElectrolytePropertySet)
        obj.add_properties(
            {
                "dh_rxn": {"method": "_dh_rxn"},
                "k_eq": {"method": "_k_eq"},
                "log_k_eq": {"method": "_log_k_eq"},
                "k_rxn": {"method": "_k_rxn"},
                "reaction_rate": {"method": "_reaction_rate"},
            }
        )


class _GenericReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(blk, outlvl=idaeslog.NOTSET, **kwargs):
        """
        Initialization routine for reaction package.

        Keyword Arguments:
            outlvl : sets output level of initialization routine

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="properties")
        init_log.info("Initialization Complete.")


@declare_process_block_class("GenericReactionBlock", block_class=_GenericReactionBlock)
class GenericReactionBlockData(ReactionBlockDataBase):
    """
    Modular Reaction Block class.
    """

    def build(self):
        # TODO: Need a different error here
        super(GenericReactionBlockData, self).build()

        if self.config.has_equilibrium:
            if len(self.params.config.equilibrium_reactions) == 0:
                raise PropertyPackageError(
                    "{} Generic Reaction Block was set to include equilibrium "
                    "reactions, however no equilibrium reactions were "
                    "defined. Either set has_equilibrium to be False, or "
                    "include equilibrium reactions in the package definition.".format(
                        self.name
                    )
                )
            self._equilibrium_constraint()

    def calculate_scaling_factors(self):
        # Get default scale factors and do calculations from base classes
        super().calculate_scaling_factors()

        # Lock attribute creation to avoid hasattr triggering builds
        self._lock_attribute_creation = True

        # Due to the exponential relationship between most reaction properties
        # and temperature, it is very hard to calculate good scaling factors
        # from order-of-magnitude guesses. Thus ,reaction scaling will always
        # require a lot of user input. Here we will calculate scaling factors
        # for those properties that have non-exponential relationships to T.
        if hasattr(self, "dh_rxn"):
            for r, v in self.dh_rxn.items():
                if iscale.get_scaling_factor(v) is None:
                    rblock = getattr(self.params, "reaction_" + r)
                    try:
                        carg = self.params.config.rate_reactions[r]
                    except (AttributeError, KeyError):
                        carg = self.params.config.equilibrium_reactions[r]
                    sf = carg["heat_of_reaction"].calculate_scaling_factors(
                        self, rblock
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "k_eq"):
            for r, v in self.k_eq.items():
                if iscale.get_scaling_factor(v) is None:
                    rblock = getattr(self.params, "reaction_" + r)
                    carg = self.params.config.equilibrium_reactions[r]
                    sf = carg["equilibrium_constant"].calculate_scaling_factors(
                        self, rblock
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "log_k_eq"):
            for r, v in self.log_k_eq.items():
                if iscale.get_scaling_factor(v) is None:
                    rblock = getattr(self.params, "reaction_" + r)
                    carg = self.params.config.equilibrium_reactions[r]
                    sf = carg["equilibrium_constant"].calculate_scaling_factors(
                        self, rblock
                    )
                    iscale.set_scaling_factor(v, log(sf))

        if hasattr(self, "equilibrium_constraint"):
            for r, v in self.equilibrium_constraint.items():
                carg = self.params.config.equilibrium_reactions[r]
                if iscale.get_scaling_factor(v) is None:
                    if carg["equilibrium_form"].__name__.startswith("log_"):
                        # Log form constraint
                        sf = iscale.get_scaling_factor(
                            self.log_k_eq[r], default=1, warning=True
                        )
                    else:
                        sf = iscale.get_scaling_factor(
                            self.k_eq[r], default=1, warning=True
                        )

                sf_const = carg["equilibrium_form"].calculate_scaling_factors(self, sf)

                iscale.constraint_scaling_transform(v, sf_const, overwrite=False)

        # Unlock attribute creation when done
        self._lock_attribute_creation = False

    def _dh_rxn(self):
        def dh_rule(b, r):
            rblock = getattr(b.params, "reaction_" + r)
            try:
                carg = b.params.config.rate_reactions[r]
            except (AttributeError, KeyError):
                carg = b.params.config.equilibrium_reactions[r]
            return carg["heat_of_reaction"].return_expression(
                b, rblock, r, b.state_ref.temperature
            )

        self.dh_rxn = Expression(
            self.params.reaction_idx, doc="Specific heat of reaction", rule=dh_rule
        )

    def _k_rxn(self):
        def krxn_rule(b, r):
            rblock = getattr(b.params, "reaction_" + r)

            carg = b.params.config.rate_reactions[r]

            return carg["rate_constant"].return_expression(
                b, rblock, r, b.state_ref.temperature
            )

        self.k_rxn = Expression(
            self.params.rate_reaction_idx, doc="Reaction rate constant", rule=krxn_rule
        )

    def _reaction_rate(self):
        def rate_rule(b, r):
            rblock = getattr(b.params, "reaction_" + r)

            carg = b.params.config.rate_reactions[r]

            if carg["rate_form"] is None:
                raise ConfigurationError(
                    f"{b.name} Generic Reaction {r} was not provided with a "
                    f"rate_form configuration argument."
                )

            return carg["rate_form"].return_expression(
                b, rblock, r, b.state_ref.temperature
            )

        self.reaction_rate = Expression(
            self.params.rate_reaction_idx, doc="Reaction rate", rule=rate_rule
        )

    def _k_eq(self):
        def keq_rule(b, r):
            rblock = getattr(b.params, "reaction_" + r)

            carg = b.params.config.equilibrium_reactions[r]

            return carg["equilibrium_constant"].return_expression(
                b, rblock, r, b.state_ref.temperature
            )

        self.k_eq = Expression(
            self.params.equilibrium_reaction_idx,
            doc="Equilibrium constant",
            rule=keq_rule,
        )

    def _log_k_eq(self):
        def log_keq_rule(b, r):
            rblock = getattr(b.params, "reaction_" + r)

            carg = b.params.config.equilibrium_reactions[r]

            return carg["equilibrium_constant"].return_log_expression(
                b, rblock, r, b.state_ref.temperature
            )

        self.log_k_eq = Var(
            self.params.equilibrium_reaction_idx,
            initialize=1,
            doc="Log of equilibrium constant",
            units=pyunits.dimensionless,
        )

        self.log_k_eq_constraint = Constraint(
            self.params.equilibrium_reaction_idx,
            rule=log_keq_rule,
            doc="Constraint for log of K_eq",
        )

    def _equilibrium_constraint(self):
        def equil_rule(b, r):
            rblock = getattr(b.params, "reaction_" + r)

            carg = b.params.config.equilibrium_reactions[r]

            return carg["equilibrium_form"].return_expression(
                b, rblock, r, b.state_ref.temperature
            )

        self.equilibrium_constraint = Constraint(
            self.params.equilibrium_reaction_idx,
            doc="Equilibrium constraint",
            rule=equil_rule,
        )

    def get_reaction_rate_basis(b):
        return b.params.config.reaction_basis



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for solid-liquid separator unit model.
Authors: Andrew Lee
"""

import pytest

from pyomo.environ import check_optimal_termination, ConcreteModel, units, value, Var
from pyomo.network import Port
from pyomo.util.check_units import assert_units_consistent

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    MomentumBalanceType,
)
from idaes.models.unit_models.solid_liquid import SLSeparator
from idaes.models.unit_models.separator import (
    Separator,
    SeparatorData,
    EnergySplittingType,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)
from idaes.core.solvers import get_solver
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
class TestSLSeparatorBasic:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = SLSeparator(
            solid_property_package=m.fs.properties,
            liquid_property_package=m.fs.properties,
        )

        m.fs.unit.solid_inlet.flow_vol.fix(10)
        m.fs.unit.solid_inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.solid_inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.solid_inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.solid_inlet.conc_mol_comp[0, "SodiumAcetate"].fix(1e-6)
        m.fs.unit.solid_inlet.conc_mol_comp[0, "Ethanol"].fix(1e-6)
        m.fs.unit.solid_inlet.temperature.fix(303.15)
        m.fs.unit.solid_inlet.pressure.fix(101325.0)

        m.fs.unit.liquid_inlet.flow_vol.fix(20)
        m.fs.unit.liquid_inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.liquid_inlet.conc_mol_comp[0, "NaOH"].fix(1e-6)
        m.fs.unit.liquid_inlet.conc_mol_comp[0, "EthylAcetate"].fix(1e-6)
        m.fs.unit.liquid_inlet.conc_mol_comp[0, "SodiumAcetate"].fix(50.0)
        m.fs.unit.liquid_inlet.conc_mol_comp[0, "Ethanol"].fix(50.0)
        m.fs.unit.liquid_inlet.temperature.fix(320)
        m.fs.unit.liquid_inlet.pressure.fix(2e5)

        m.fs.unit.liquid_recovery.fix(0.7)

        return m

    @pytest.mark.unit
    def test_config(self, model):
        # Check unit config arguments
        assert len(model.fs.unit.config) == 9

        assert not model.fs.unit.config.dynamic
        assert not model.fs.unit.config.has_holdup
        assert (
            model.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
        )
        assert (
            model.fs.unit.config.energy_split_basis
            == EnergySplittingType.equal_temperature
        )
        assert (
            model.fs.unit.config.momentum_balance_type
            == MomentumBalanceType.pressureTotal
        )

        assert model.fs.unit.config.solid_property_package is model.fs.properties
        assert model.fs.unit.config.liquid_property_package is model.fs.properties

        assert model.fs.unit.default_initializer is BlockTriangularizationInitializer

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):

        assert isinstance(model.fs.unit.solid_inlet, Port)
        assert len(model.fs.unit.solid_inlet.vars) == 4
        assert hasattr(model.fs.unit.solid_inlet, "flow_vol")
        assert hasattr(model.fs.unit.solid_inlet, "conc_mol_comp")
        assert hasattr(model.fs.unit.solid_inlet, "temperature")
        assert hasattr(model.fs.unit.solid_inlet, "pressure")

        assert isinstance(model.fs.unit.solid_outlet, Port)
        assert len(model.fs.unit.solid_outlet.vars) == 4
        assert hasattr(model.fs.unit.solid_outlet, "flow_vol")
        assert hasattr(model.fs.unit.solid_outlet, "conc_mol_comp")
        assert hasattr(model.fs.unit.solid_outlet, "temperature")
        assert hasattr(model.fs.unit.solid_outlet, "pressure")

        assert isinstance(model.fs.unit.liquid_inlet, Port)
        assert len(model.fs.unit.liquid_inlet.vars) == 4
        assert hasattr(model.fs.unit.liquid_inlet, "flow_vol")
        assert hasattr(model.fs.unit.liquid_inlet, "conc_mol_comp")
        assert hasattr(model.fs.unit.liquid_inlet, "temperature")
        assert hasattr(model.fs.unit.liquid_inlet, "pressure")

        assert isinstance(model.fs.unit.retained_liquid_outlet, Port)
        assert len(model.fs.unit.retained_liquid_outlet.vars) == 4
        assert hasattr(model.fs.unit.retained_liquid_outlet, "flow_vol")
        assert hasattr(model.fs.unit.retained_liquid_outlet, "conc_mol_comp")
        assert hasattr(model.fs.unit.retained_liquid_outlet, "temperature")
        assert hasattr(model.fs.unit.retained_liquid_outlet, "pressure")

        assert isinstance(model.fs.unit.recovered_liquid_outlet, Port)
        assert len(model.fs.unit.recovered_liquid_outlet.vars) == 4
        assert hasattr(model.fs.unit.recovered_liquid_outlet, "flow_vol")
        assert hasattr(model.fs.unit.recovered_liquid_outlet, "conc_mol_comp")
        assert hasattr(model.fs.unit.recovered_liquid_outlet, "temperature")
        assert hasattr(model.fs.unit.recovered_liquid_outlet, "pressure")

        assert isinstance(model.fs.unit.split, SeparatorData)
        assert isinstance(model.fs.unit.liquid_recovery, Var)

        assert number_variables(model) == 34
        assert number_total_constraints(model) == 17
        # These are the solid properties, as they do not appear in constraints
        assert number_unused_variables(model) == 8

    @pytest.mark.component
    def test_units(self, model):
        assert_units_consistent(model)

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_structural_issues(self, model):
        dt = DiagnosticsToolbox(model)
        dt.assert_no_structural_warnings()

    @pytest.mark.component
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(constraint_tolerance=2e-5)
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model, tee=True)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, model):
        dt = DiagnosticsToolbox(model)
        dt.assert_no_numerical_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # Solid outlet
        assert pytest.approx(101325.0, rel=1e-8) == value(
            model.fs.unit.solid_outlet.pressure[0]
        )
        assert pytest.approx(303.15, rel=1e-8) == value(
            model.fs.unit.solid_outlet.temperature[0]
        )
        assert pytest.approx(100, rel=1e-8) == value(
            model.fs.unit.solid_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(100, rel=1e-8) == value(
            model.fs.unit.solid_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(1e-6, rel=1e-8) == value(
            model.fs.unit.solid_outlet.conc_mol_comp[0, "Ethanol"]
        )
        assert pytest.approx(1e-6, rel=1e-8) == value(
            model.fs.unit.solid_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(10, rel=1e-8) == value(
            model.fs.unit.solid_outlet.flow_vol[0]
        )

        # Retained liquid
        assert pytest.approx(2e5, rel=1e-8) == value(
            model.fs.unit.retained_liquid_outlet.pressure[0]
        )
        assert pytest.approx(320, rel=1e-8) == value(
            model.fs.unit.retained_liquid_outlet.temperature[0]
        )
        assert pytest.approx(1e-6, rel=1e-8) == value(
            model.fs.unit.retained_liquid_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(1e-6, rel=1e-8) == value(
            model.fs.unit.retained_liquid_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(50, rel=1e-8) == value(
            model.fs.unit.retained_liquid_outlet.conc_mol_comp[0, "Ethanol"]
        )
        assert pytest.approx(50, rel=1e-8) == value(
            model.fs.unit.retained_liquid_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(20 * 0.3, rel=1e-8) == value(
            model.fs.unit.retained_liquid_outlet.flow_vol[0]
        )

        # Recovered liquid
        assert pytest.approx(2e5, rel=1e-8) == value(
            model.fs.unit.recovered_liquid_outlet.pressure[0]
        )
        assert pytest.approx(320, rel=1e-8) == value(
            model.fs.unit.recovered_liquid_outlet.temperature[0]
        )
        assert pytest.approx(1e-6, rel=1e-8) == value(
            model.fs.unit.recovered_liquid_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(1e-6, rel=1e-8) == value(
            model.fs.unit.recovered_liquid_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(50, rel=1e-8) == value(
            model.fs.unit.recovered_liquid_outlet.conc_mol_comp[0, "Ethanol"]
        )
        assert pytest.approx(50, rel=1e-8) == value(
            model.fs.unit.recovered_liquid_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(20 * 0.7, rel=1e-8) == value(
            model.fs.unit.recovered_liquid_outlet.flow_vol[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solid_conservation(self, model):
        assert (
            abs(
                value(
                    model.fs.unit.solid_inlet.flow_vol[0]
                    - model.fs.unit.solid_outlet.flow_vol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    model.fs.unit.solid_inlet.flow_vol[0]
                    * sum(
                        model.fs.unit.solid_inlet.conc_mol_comp[0, j]
                        for j in model.fs.properties.component_list
                    )
                    - model.fs.unit.solid_outlet.flow_vol[0]
                    * sum(
                        model.fs.unit.solid_outlet.conc_mol_comp[0, j]
                        for j in model.fs.properties.component_list
                    )
                )
            )
            <= 1e-6
        )

        assert (
            abs(
                value(
                    (
                        model.fs.unit.solid_inlet.flow_vol[0]
                        * model.fs.properties.dens_mol
                        * model.fs.properties.cp_mol
                        * (
                            model.fs.unit.solid_inlet.temperature[0]
                            - model.fs.properties.temperature_ref
                        )
                    )
                    - (
                        model.fs.unit.solid_outlet.flow_vol[0]
                        * model.fs.properties.dens_mol
                        * model.fs.properties.cp_mol
                        * (
                            model.fs.unit.solid_outlet.temperature[0]
                            - model.fs.properties.temperature_ref
                        )
                    )
                )
            )
            <= 1e-3
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_liquid_conservation(self, model):
        assert (
            abs(
                value(
                    model.fs.unit.liquid_inlet.flow_vol[0]
                    - model.fs.unit.retained_liquid_outlet.flow_vol[0]
                    - model.fs.unit.recovered_liquid_outlet.flow_vol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    model.fs.unit.liquid_inlet.flow_vol[0]
                    * sum(
                        model.fs.unit.liquid_inlet.conc_mol_comp[0, j]
                        for j in model.fs.properties.component_list
                    )
                    - model.fs.unit.retained_liquid_outlet.flow_vol[0]
                    * sum(
                        model.fs.unit.retained_liquid_outlet.conc_mol_comp[0, j]
                        for j in model.fs.properties.component_list
                    )
                    - model.fs.unit.recovered_liquid_outlet.flow_vol[0]
                    * sum(
                        model.fs.unit.recovered_liquid_outlet.conc_mol_comp[0, j]
                        for j in model.fs.properties.component_list
                    )
                )
            )
            <= 1e-6
        )

        assert (
            abs(
                value(
                    (
                        model.fs.unit.liquid_inlet.flow_vol[0]
                        * model.fs.properties.dens_mol
                        * model.fs.properties.cp_mol
                        * (
                            model.fs.unit.liquid_inlet.temperature[0]
                            - model.fs.properties.temperature_ref
                        )
                    )
                    - (
                        model.fs.unit.retained_liquid_outlet.flow_vol[0]
                        * model.fs.properties.dens_mol
                        * model.fs.properties.cp_mol
                        * (
                            model.fs.unit.retained_liquid_outlet.temperature[0]
                            - model.fs.properties.temperature_ref
                        )
                    )
                    - (
                        model.fs.unit.recovered_liquid_outlet.flow_vol[0]
                        * model.fs.properties.dens_mol
                        * model.fs.properties.cp_mol
                        * (
                            model.fs.unit.recovered_liquid_outlet.temperature[0]
                            - model.fs.properties.temperature_ref
                        )
                    )
                )
            )
            <= 1e-3
        )

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, model):
        perf_dict = model.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Liquid Recovery": model.fs.unit.liquid_recovery[0],
            }
        }

    @pytest.mark.ui
    @pytest.mark.component
    def test_get_stream_table_contents(self, model):
        stable = model.fs.unit._get_stream_table_contents()

        expected = {
            "Units": {
                "Volumetric Flowrate": getattr(units.pint_registry, "m**3/second"),
                "Molar Concentration H2O": getattr(units.pint_registry, "mole/m**3"),
                "Molar Concentration NaOH": getattr(units.pint_registry, "mole/m**3"),
                "Molar Concentration EthylAcetate": getattr(
                    units.pint_registry, "mole/m**3"
                ),
                "Molar Concentration SodiumAcetate": getattr(
                    units.pint_registry, "mole/m**3"
                ),
                "Molar Concentration Ethanol": getattr(
                    units.pint_registry, "mole/m**3"
                ),
                "Temperature": getattr(units.pint_registry, "K"),
                "Pressure": getattr(units.pint_registry, "Pa"),
            },
            "Solid Inlet": {
                "Volumetric Flowrate": pytest.approx(10, rel=1e-4),
                "Molar Concentration H2O": pytest.approx(5.5388e4, rel=1e-4),
                "Molar Concentration NaOH": pytest.approx(100, rel=1e-4),
                "Molar Concentration EthylAcetate": pytest.approx(100, rel=1e-4),
                "Molar Concentration SodiumAcetate": pytest.approx(1e-6, abs=1e-4),
                "Molar Concentration Ethanol": pytest.approx(1e-6, rel=1e-4),
                "Temperature": pytest.approx(303.15, rel=1e-4),
                "Pressure": pytest.approx(101325, rel=1e-4),
            },
            "Liquid Inlet": {
                "Volumetric Flowrate": pytest.approx(20, rel=1e-4),
                "Molar Concentration H2O": pytest.approx(5.5388e4, rel=1e-4),
                "Molar Concentration NaOH": pytest.approx(1e-6, rel=1e-4),
                "Molar Concentration EthylAcetate": pytest.approx(1e-6, rel=1e-4),
                "Molar Concentration SodiumAcetate": pytest.approx(50, rel=1e-4),
                "Molar Concentration Ethanol": pytest.approx(50, rel=1e-4),
                "Temperature": pytest.approx(320, rel=1e-4),
                "Pressure": pytest.approx(2e5, rel=1e-4),
            },
            "Solid Outlet": {
                "Volumetric Flowrate": pytest.approx(10, rel=1e-4),
                "Molar Concentration H2O": pytest.approx(5.5388e4, rel=1e-4),
                "Molar Concentration NaOH": pytest.approx(100, rel=1e-4),
                "Molar Concentration EthylAcetate": pytest.approx(100, rel=1e-4),
                "Molar Concentration SodiumAcetate": pytest.approx(1e-6, rel=1e-4),
                "Molar Concentration Ethanol": pytest.approx(1e-6, rel=1e-4),
                "Temperature": pytest.approx(303.15, rel=1e-4),
                "Pressure": pytest.approx(101325, rel=1e-4),
            },
            "Liquid in Solids Outlet": {
                "Volumetric Flowrate": pytest.approx(20 * 0.3, rel=1e-4),
                "Molar Concentration H2O": pytest.approx(5.5388e4, rel=1e-4),
                "Molar Concentration NaOH": pytest.approx(1e-6, rel=1e-4),
                "Molar Concentration EthylAcetate": pytest.approx(1e-6, rel=1e-4),
                "Molar Concentration SodiumAcetate": pytest.approx(50, rel=1e-4),
                "Molar Concentration Ethanol": pytest.approx(50, rel=1e-4),
                "Temperature": pytest.approx(320, rel=1e-4),
                "Pressure": pytest.approx(2e5, rel=1e-4),
            },
            "Recovered Liquid Outlet": {
                "Volumetric Flowrate": pytest.approx(20 * 0.7, rel=1e-4),
                "Molar Concentration H2O": pytest.approx(5.5388e4, rel=1e-4),
                "Molar Concentration NaOH": pytest.approx(1e-6, rel=1e-4),
                "Molar Concentration EthylAcetate": pytest.approx(1e-6, rel=1e-4),
                "Molar Concentration SodiumAcetate": pytest.approx(50, rel=1e-4),
                "Molar Concentration Ethanol": pytest.approx(50, rel=1e-4),
                "Temperature": pytest.approx(320, rel=1e-4),
                "Pressure": pytest.approx(2e5, rel=1e-4),
            },
        }

        assert stable.to_dict() == expected



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests ProcessBlock and ProcessBlockData.

Author: John Eslick
"""
import pytest
from pyomo.environ import ConcreteModel, Var, value
from pyomo.common.config import ConfigValue
from idaes.core import ProcessBlockData, declare_process_block_class


@declare_process_block_class("MyBlock")
class MyBlockData(ProcessBlockData):
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare("xinit", ConfigValue(default=1001, domain=float))
    CONFIG.declare("yinit", ConfigValue(default=1002, domain=float))

    def build(self):
        super(MyBlockData, self).build()
        self.x = Var(initialize=self.config.xinit)
        self.y = Var(initialize=self.config.yinit)


class TestProcessBlock(object):
    @pytest.mark.unit
    def test_scalar_noargs(self):
        m = ConcreteModel()
        m.b = MyBlock()
        assert isinstance(m.b.x, Var)
        assert isinstance(m.b.y, Var)
        assert value(m.b.x) == 1001
        assert value(m.b.y) == 1002
        assert m.b.process_block_class() is MyBlock
        assert m.b.parent_component().process_block_class() is MyBlock

    @pytest.mark.unit
    def test_vec_noargs(self):
        m = ConcreteModel()
        m.b = MyBlock([1, 2, 3])
        assert isinstance(m.b[1].x, Var)
        assert isinstance(m.b[1].y, Var)
        assert isinstance(m.b[2].x, Var)
        assert isinstance(m.b[2].y, Var)
        assert isinstance(m.b[3].x, Var)
        assert isinstance(m.b[3].y, Var)
        assert value(m.b[1].x) == 1001
        assert value(m.b[1].y) == 1002
        assert value(m.b[2].x) == 1001
        assert value(m.b[2].y) == 1002
        assert value(m.b[3].x) == 1001
        assert value(m.b[3].y) == 1002
        assert m.b[2].parent_component().process_block_class() is MyBlock
        assert m.b.process_block_class() is MyBlock

    @pytest.mark.unit
    def test_scalar_args1(self):
        m = ConcreteModel()
        m.b = MyBlock(xinit=1, yinit=2)
        assert isinstance(m.b.x, Var)
        assert isinstance(m.b.y, Var)
        assert value(m.b.x) == 1
        assert value(m.b.y) == 2
        assert m.b.process_block_class() is MyBlock
        assert m.b.parent_component().process_block_class() is MyBlock

    @pytest.mark.unit
    def test_scalar_args2(self):
        m = ConcreteModel()
        m.b = MyBlock(initialize={None: {"xinit": 1, "yinit": 2}})
        assert isinstance(m.b.x, Var)
        assert isinstance(m.b.y, Var)
        assert value(m.b.x) == 1
        assert value(m.b.y) == 2
        assert m.b.process_block_class() is MyBlock
        assert m.b.parent_component().process_block_class() is MyBlock

    @pytest.mark.unit
    def test_vec_args(self):
        m = ConcreteModel()
        m.b = MyBlock(
            [1, 2, 3], xinit=1, yinit=2, initialize={2: {"xinit": 2001, "yinit": 2002}}
        )
        assert isinstance(m.b[1].x, Var)
        assert isinstance(m.b[1].y, Var)
        assert isinstance(m.b[2].x, Var)
        assert isinstance(m.b[2].y, Var)
        assert isinstance(m.b[3].x, Var)
        assert isinstance(m.b[3].y, Var)
        assert m.b[1].index() == 1
        assert value(m.b[1].x) == 1
        assert value(m.b[1].y) == 2
        assert value(m.b[2].x) == 2001
        assert value(m.b[2].y) == 2002
        assert value(m.b[3].x) == 1
        assert value(m.b[3].y) == 2
        assert m.b[2].parent_component().process_block_class() is MyBlock
        assert m.b.process_block_class() is MyBlock

    @pytest.mark.unit
    def test_user_map(self):
        m = ConcreteModel()

        def new_imap(i):
            if i > 0 and i < 4:
                return 1
            else:
                return i

        m.b = MyBlock(
            [0, 1, 2, 3, 4],
            idx_map=new_imap,
            initialize={
                0: {"xinit": 2001, "yinit": 2002},
                1: {"xinit": 5001, "yinit": 5002},
                2: {"xinit": 6001, "yinit": 6002},
                4: {"xinit": 7001, "yinit": 7002},
            },
        )
        assert value(m.b[0].x) == 2001
        assert value(m.b[0].y) == 2002
        assert value(m.b[1].x) == 5001
        assert value(m.b[1].y) == 5002
        assert value(m.b[2].x) == 5001  # although this index (2) is in initialize
        assert value(m.b[2].y) == 5002  # the idx_map function maps it to 1
        assert value(m.b[3].x) == 5001
        assert value(m.b[3].y) == 5002
        assert value(m.b[4].x) == 7001
        assert value(m.b[4].y) == 7002
        assert m.b[2].parent_component().process_block_class() is MyBlock
        assert m.b.process_block_class() is MyBlock



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring

import collections

from pyomo.core.base.block import declare_custom_block, BlockData
from pyomo.environ import Var, Set

import idaes.logger as idaeslog

__author__ = "Carl Laird, Andrew Lee"
_log = idaeslog.getLogger(__name__)


@declare_custom_block(name="SurrogateBlock")
class SurrogateBlockData(BlockData):
    # pylint: disable=W0235
    def __init__(self, component):
        """
        This custom block is used for importing surrogates into IDAES

        Example usage without specifying inputs and outputs:
        > # Load the surrogate object
        > alamo_surrogate = AlamoSurrogate.load_from_file('alamo_reformer.json')
        >
        > # create the Pyomo/IDAES model
        > m = ConcreteModel()
        >
        > # create the Pyomo/IDAES block that corresponds to the surrogate
        > m.reformer = SurrogateBlock()
        > m.reformer.build_model(alamo_surrogate)
        >
        > # maximize the concentration of H2
        > m.obj = Objective(expr=m.reformer.outputs['H2'], sense=maximize)
        >
        > # constrain the outlet concentration of N2
        > m.reformer.outputs['N2'].setub(0.34)
        >
        > status = SolverFactory('ipopt').solve(m, tee=True)
        > m.reformer.display()

        Note that you can also provide variables from your IDAES model as inputs
        and outputs. For example:
        > m.reformer.build_model(alamo_surrogate, \
        >     input_vars=[m.fs.reformer_duty, m.fs.product.inlet.flow_mol],
        >     output_vars=[m.fs.total_cost])

        See the specific methods for more documentation.
        """
        super(SurrogateBlockData, self).__init__(component)

    def build_model(
        self,
        surrogate_object,
        input_vars=None,
        output_vars=None,
        use_surrogate_bounds=True,
        **kwargs,
    ):
        """
        Build an EO model of the surrogate on the block. This method will build the
        necessary equations, and will construct inputs and outputs if necessary (i.e.,
        they are not provided)

        Args:
           surrogate_object: class derived from SurrogateBase (e.g., AlamoSurrogate)
              This is the surrogate object that is used to build the equations.
           input_vars : None or list of scalar or indexed variables
              If None, then the input variables are automatically created as block.inputs with keys
              coming from the input_labels of the surrogate_object. If a list is provided, that list
              can contain Var objects (indexed, scalar, or VarData) where the order must match the order
              of the inputs on the surrogate model. If input_vars are provided (not None), then
              block.inputs will not be created on the model.
           output_vars : None or list of scalar or indexed variables
              If None, then the output variables are automatically created as block.outputs with keys
              coming from the output_labels of the surrogate_object. If a list is provided, that list
              can contain Var objects (indexed, scalar, or VarData) where the order must match the order
              of the outputs on the surrogate model.  If output_vars are provided (not None), then
              block.outputs will not be created on the model.
           use_surrogate_bounds : bool
              If True, then the bounds on the input variables are updated using the bounds
              in the surrogate object. If False, then the bounds on the input variables are not overwritten.

        Any additional keyword arguments are passed to the populate_block method of the underlying surrogate object.
        Please see the documentation for the specific surrogate object for details.
        """
        self._setup_inputs_outputs(
            n_inputs=surrogate_object.n_inputs(),
            n_outputs=surrogate_object.n_outputs(),
            input_vars=input_vars,
            input_labels=surrogate_object.input_labels(),
            output_vars=output_vars,
            output_labels=surrogate_object.output_labels(),
        )

        # set the input bounds if desired
        input_bounds = surrogate_object.input_bounds()
        input_vars_as_dict = self.input_vars_as_dict()
        if use_surrogate_bounds is True and input_bounds is not None:
            for k, bnd in input_bounds.items():
                v = input_vars_as_dict[k]
                lb = bnd[0]
                if v.lb is not None:
                    lb = max(lb, v.lb)
                ub = bnd[1]
                if v.ub is not None:
                    ub = min(ub, v.ub)
                _log.debug(f"Setting bound of {v} to ({lb}, {ub}).")
                v.setlb(lb)
                v.setub(ub)

        # call populate block to fill-in the constraints
        surrogate_object.populate_block(self, additional_options=kwargs)

        # test that kwargs is empty
        # derived classes should call .pop when they use a keyword argument
        if len(kwargs) > 0:
            raise ValueError(
                "Error in keyword arguments passed to build_model."
                " The following arguments were not used: {}".format([k for k in kwargs])
            )

    def _setup_inputs_outputs(
        self,
        n_inputs,
        n_outputs,
        input_vars=None,
        input_labels=None,
        output_vars=None,
        output_labels=None,
    ):
        if n_inputs < 1 or n_outputs < 1:
            raise ValueError(
                "Attempting to create a Surrogate block with no inputs "
                "and/or no outputs. A SurrogateBlock must have at least "
                "one input and output"
            )

        # copy or create the labels
        if input_labels is None:
            self._input_labels = list(range(n_inputs))
        else:
            self._input_labels = list(input_labels)
            if len(self._input_labels) != n_inputs:
                raise ValueError(
                    "Specifying input_labels to a SurrogateBlock, but the "
                    "length does not match n_inputs"
                )

        # create or extract the variables
        if input_vars is None:
            # we need to create our own inputs
            self.inputs_set = Set(initialize=self._input_labels, ordered=True)
            self.inputs = Var(self.inputs_set, initialize=0)
            self._input_vars = list(self.inputs.values())
        else:
            # we were provided vars
            self._input_vars = _extract_var_data(input_vars)
            if len(self._input_vars) != n_inputs:
                raise ValueError(
                    "Specifying input_vars to a SurrogateBlock, but the"
                    " length of the input_vars (after extracting all"
                    " the VarData objects) does not match n_inputs"
                )

        if output_labels is None:
            self._output_labels = list(range(n_outputs))
        else:
            self._output_labels = list(output_labels)
            if len(self._output_labels) != n_outputs:
                raise ValueError(
                    "Specifying output_labels to a SurrogateBlock, but the "
                    "length does not match n_outputs"
                )

        # create or extract the output variables
        if output_vars is None:
            # we need to create our own outputs
            self.outputs_set = Set(initialize=self._output_labels, ordered=True)
            self.outputs = Var(self.outputs_set, initialize=0)
            self._output_vars = list(self.outputs.values())
        else:
            # we were provided vars
            self._output_vars = _extract_var_data(output_vars)
            if len(self._output_vars) != n_outputs:
                raise ValueError(
                    "Specifying output_vars to a SurrogateBlock, but the"
                    " length of the output_vars (after extracting all"
                    " the VarData objects) does not match n_outputs"
                )

    def _input_vars_as_list(self):
        return self._input_vars

    def _output_vars_as_list(self):
        return self._output_vars

    def input_vars_as_dict(self):
        """
        Returns a dictionary of the input variables (VarData objects) with
        the labels as the keys
        """
        return {self._input_labels[i]: v for i, v in enumerate(self._input_vars)}

    def output_vars_as_dict(self):
        """
        Returns a dictionary of the output variables (VarData objects) with
        the labels as the keys
        """
        return {self._output_labels[i]: v for i, v in enumerate(self._output_vars)}


def _extract_var_data_gen(_vars):
    if _vars is None:
        return
    elif getattr(_vars, "ctype", None) is Var:
        if _vars.is_indexed():
            if not _vars.index_set().isordered():
                raise ValueError(
                    f"Expected IndexedVar: {_vars} to be indexed over an ordered set."
                )
            yield from _vars.values()
        else:
            yield _vars
    elif isinstance(_vars, collections.abc.Sequence):
        for v in _vars:
            yield from _extract_var_data_gen(v)
    else:
        raise ValueError(f"Unknown variable type of {type(_vars)} for {_vars}")


def _extract_var_data(_vars):
    if _vars is None:
        return None
    return list(_extract_var_data_gen(_vars))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for calculating heat of reaction
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import Var, value

from idaes.core import MaterialFlowBasis
from idaes.core.util.misc import set_param_from_config


# -----------------------------------------------------------------------------
# Constant dh_rxn
class constant_dh_rxn:
    """Methods for constant heat of reaction."""

    @staticmethod
    def build_parameters(rblock, config):
        units = rblock.parent_block().get_metadata().derived_units

        rbasis = rblock.parent_block().config.reaction_basis
        if rbasis == MaterialFlowBasis.molar:
            basis = "mole"
        elif rbasis == MaterialFlowBasis.mass:
            basis = "mass"

        rblock.dh_rxn_ref = Var(
            doc="Specific heat of reaction at reference state",
            units=units["energy_" + basis],
        )

        set_param_from_config(rblock, param="dh_rxn_ref", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        return rblock.dh_rxn_ref

    @staticmethod
    def calculate_scaling_factors(b, rblock):
        v = abs(value(rblock.dh_rxn_ref))

        # Need to make sure dh_rxn is not 0 to avoid division by 0
        if v != 0:
            return 1 / abs(value(rblock.dh_rxn_ref))
        else:
            return 1
